# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
from pathlib import Path
from types import SimpleNamespace
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Pattern,
    Sequence,
    Set,
    Union,
    cast,
)

from playwright._impl._api_structures import (
    Cookie,
    Geolocation,
    SetCookieParam,
    StorageState,
)
from playwright._impl._artifact import Artifact
from playwright._impl._cdp_session import CDPSession
from playwright._impl._clock import Clock
from playwright._impl._connection import (
    ChannelOwner,
    from_channel,
    from_nullable_channel,
)
from playwright._impl._console_message import ConsoleMessage
from playwright._impl._dialog import Dialog
from playwright._impl._errors import Error, TargetClosedError
from playwright._impl._event_context_manager import EventContextManagerImpl
from playwright._impl._fetch import APIRequestContext
from playwright._impl._frame import Frame
from playwright._impl._har_router import HarRouter
from playwright._impl._helper import (
    HarContentPolicy,
    HarMode,
    HarRecordingMetadata,
    RouteFromHarNotFoundPolicy,
    RouteHandler,
    RouteHandlerCallback,
    TimeoutSettings,
    URLMatch,
    WebSocketRouteHandlerCallback,
    async_readfile,
    async_writefile,
    locals_to_params,
    parse_error,
    to_impl,
)
from playwright._impl._network import (
    Request,
    Response,
    Route,
    WebSocketRoute,
    WebSocketRouteHandler,
    serialize_headers,
)
from playwright._impl._page import BindingCall, Page, Worker
from playwright._impl._str_utils import escape_regex_flags
from playwright._impl._tracing import Tracing
from playwright._impl._waiter import Waiter
from playwright._impl._web_error import WebError

if TYPE_CHECKING:  # pragma: no cover
    from playwright._impl._browser import Browser


class BrowserContext(ChannelOwner):
    Events = SimpleNamespace(
        BackgroundPage="backgroundpage",
        Close="close",
        Console="console",
        Dialog="dialog",
        Page="page",
        WebError="weberror",
        ServiceWorker="serviceworker",
        Request="request",
        Response="response",
        RequestFailed="requestfailed",
        RequestFinished="requestfinished",
    )

    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)
        # Browser is null for browser contexts created outside of normal browser, e.g. android or electron.
        # circular import workaround:
        self._browser: Optional["Browser"] = None
        if parent.__class__.__name__ == "Browser":
            self._browser = cast("Browser", parent)
        self._pages: List[Page] = []
        self._routes: List[RouteHandler] = []
        self._web_socket_routes: List[WebSocketRouteHandler] = []
        self._bindings: Dict[str, Any] = {}
        self._timeout_settings = TimeoutSettings(None)
        self._owner_page: Optional[Page] = None
        self._options: Dict[str, Any] = initializer["options"]
        self._background_pages: Set[Page] = set()
        self._service_workers: Set[Worker] = set()
        self._base_url: Optional[str] = self._options.get("baseURL")
        self._videos_dir: Optional[str] = self._options.get("recordVideo")
        self._tracing = cast(Tracing, from_channel(initializer["tracing"]))
        self._har_recorders: Dict[str, HarRecordingMetadata] = {}
        self._request: APIRequestContext = from_channel(initializer["requestContext"])
        self._clock = Clock(self)
        self._channel.on(
            "bindingCall",
            lambda params: self._on_binding(from_channel(params["binding"])),
        )
        self._channel.on("close", lambda _: self._on_close())
        self._channel.on(
            "page", lambda params: self._on_page(from_channel(params["page"]))
        )
        self._channel.on(
            "route",
            lambda params: self._loop.create_task(
                self._on_route(
                    from_channel(params.get("route")),
                )
            ),
        )
        self._channel.on(
            "webSocketRoute",
            lambda params: self._loop.create_task(
                self._on_web_socket_route(
                    from_channel(params["webSocketRoute"]),
                )
            ),
        )
        self._channel.on(
            "backgroundPage",
            lambda params: self._on_background_page(from_channel(params["page"])),
        )

        self._channel.on(
            "serviceWorker",
            lambda params: self._on_service_worker(from_channel(params["worker"])),
        )
        self._channel.on(
            "console",
            lambda event: self._on_console_message(event),
        )

        self._channel.on(
            "dialog", lambda params: self._on_dialog(from_channel(params["dialog"]))
        )
        self._channel.on(
            "pageError",
            lambda params: self._on_page_error(
                parse_error(params["error"]["error"]),
                from_nullable_channel(params["page"]),
            ),
        )
        self._channel.on(
            "request",
            lambda params: self._on_request(
                from_channel(params["request"]),
                from_nullable_channel(params.get("page")),
            ),
        )
        self._channel.on(
            "response",
            lambda params: self._on_response(
                from_channel(params["response"]),
                from_nullable_channel(params.get("page")),
            ),
        )
        self._channel.on(
            "requestFailed",
            lambda params: self._on_request_failed(
                from_channel(params["request"]),
                params["responseEndTiming"],
                params.get("failureText"),
                from_nullable_channel(params.get("page")),
            ),
        )
        self._channel.on(
            "requestFinished",
            lambda params: self._on_request_finished(
                from_channel(params["request"]),
                from_nullable_channel(params.get("response")),
                params["responseEndTiming"],
                from_nullable_channel(params.get("page")),
            ),
        )
        self._closed_future: asyncio.Future = asyncio.Future()
        self.once(
            self.Events.Close, lambda context: self._closed_future.set_result(True)
        )
        self._close_reason: Optional[str] = None
        self._har_routers: List[HarRouter] = []
        self._set_event_to_subscription_mapping(
            {
                BrowserContext.Events.Console: "console",
                BrowserContext.Events.Dialog: "dialog",
                BrowserContext.Events.Request: "request",
                BrowserContext.Events.Response: "response",
                BrowserContext.Events.RequestFinished: "requestFinished",
                BrowserContext.Events.RequestFailed: "requestFailed",
            }
        )
        self._closing_or_closed = False

    def __repr__(self) -> str:
        return f"<BrowserContext browser={self.browser}>"

    def _on_page(self, page: Page) -> None:
        self._pages.append(page)
        self.emit(BrowserContext.Events.Page, page)
        if page._opener and not page._opener.is_closed():
            page._opener.emit(Page.Events.Popup, page)

    async def _on_route(self, route: Route) -> None:
        route._context = self
        page = route.request._safe_page()
        route_handlers = self._routes.copy()
        for route_handler in route_handlers:
            # If the page or the context was closed we stall all requests right away.
            if (page and page._close_was_called) or self._closing_or_closed:
                return
            if not route_handler.matches(route.request.url):
                continue
            if route_handler not in self._routes:
                continue
            if route_handler.will_expire:
                self._routes.remove(route_handler)
            try:
                handled = await route_handler.handle(route)
            finally:
                if len(self._routes) == 0:
                    asyncio.create_task(
                        self._connection.wrap_api_call(
                            lambda: self._update_interception_patterns(), True
                        )
                    )
            if handled:
                return
        try:
            # If the page is closed or unrouteAll() was called without waiting and interception disabled,
            # the method will throw an error - silence it.
            await route._inner_continue(True)
        except Exception:
            pass

    async def _on_web_socket_route(self, web_socket_route: WebSocketRoute) -> None:
        route_handler = next(
            (
                route_handler
                for route_handler in self._web_socket_routes
                if route_handler.matches(web_socket_route.url)
            ),
            None,
        )
        if route_handler:
            await route_handler.handle(web_socket_route)
        else:
            web_socket_route.connect_to_server()

    def _on_binding(self, binding_call: BindingCall) -> None:
        func = self._bindings.get(binding_call._initializer["name"])
        if func is None:
            return
        asyncio.create_task(binding_call.call(func))

    def set_default_navigation_timeout(self, timeout: float) -> None:
        return self._set_default_navigation_timeout_impl(timeout)

    def _set_default_navigation_timeout_impl(self, timeout: Optional[float]) -> None:
        self._timeout_settings.set_default_navigation_timeout(timeout)

    def set_default_timeout(self, timeout: float) -> None:
        return self._set_default_timeout_impl(timeout)

    def _set_default_timeout_impl(self, timeout: Optional[float]) -> None:
        self._timeout_settings.set_default_timeout(timeout)

    @property
    def pages(self) -> List[Page]:
        return self._pages.copy()

    @property
    def browser(self) -> Optional["Browser"]:
        return self._browser

    async def _initialize_har_from_options(
        self,
        record_har_path: Optional[Union[Path, str]],
        record_har_content: Optional[HarContentPolicy],
        record_har_omit_content: Optional[bool],
        record_har_url_filter: Optional[Union[Pattern[str], str]],
        record_har_mode: Optional[HarMode],
    ) -> None:
        if not record_har_path:
            return
        record_har_path = str(record_har_path)
        default_policy: HarContentPolicy = (
            "attach" if record_har_path.endswith(".zip") else "embed"
        )
        content_policy: HarContentPolicy = record_har_content or (
            "omit" if record_har_omit_content is True else default_policy
        )
        await self._record_into_har(
            har=record_har_path,
            page=None,
            url=record_har_url_filter,
            update_content=content_policy,
            update_mode=(record_har_mode or "full"),
        )

    async def new_page(self) -> Page:
        if self._owner_page:
            raise Error("Please use browser.new_context()")
        return from_channel(await self._channel.send("newPage", None))

    async def cookies(self, urls: Union[str, Sequence[str]] = None) -> List[Cookie]:
        if urls is None:
            urls = []
        if isinstance(urls, str):
            urls = [urls]
        return await self._channel.send("cookies", None, dict(urls=urls))

    async def add_cookies(self, cookies: Sequence[SetCookieParam]) -> None:
        await self._channel.send("addCookies", None, dict(cookies=cookies))

    async def clear_cookies(
        self,
        name: Union[str, Pattern[str]] = None,
        domain: Union[str, Pattern[str]] = None,
        path: Union[str, Pattern[str]] = None,
    ) -> None:
        await self._channel.send(
            "clearCookies",
            None,
            {
                "name": name if isinstance(name, str) else None,
                "nameRegexSource": name.pattern if isinstance(name, Pattern) else None,
                "nameRegexFlags": (
                    escape_regex_flags(name) if isinstance(name, Pattern) else None
                ),
                "domain": domain if isinstance(domain, str) else None,
                "domainRegexSource": (
                    domain.pattern if isinstance(domain, Pattern) else None
                ),
                "domainRegexFlags": (
                    escape_regex_flags(domain) if isinstance(domain, Pattern) else None
                ),
                "path": path if isinstance(path, str) else None,
                "pathRegexSource": path.pattern if isinstance(path, Pattern) else None,
                "pathRegexFlags": (
                    escape_regex_flags(path) if isinstance(path, Pattern) else None
                ),
            },
        )

    async def grant_permissions(
        self, permissions: Sequence[str], origin: str = None
    ) -> None:
        await self._channel.send("grantPermissions", None, locals_to_params(locals()))

    async def clear_permissions(self) -> None:
        await self._channel.send("clearPermissions", None)

    async def set_geolocation(self, geolocation: Geolocation = None) -> None:
        await self._channel.send("setGeolocation", None, locals_to_params(locals()))

    async def set_extra_http_headers(self, headers: Dict[str, str]) -> None:
        await self._channel.send(
            "setExtraHTTPHeaders", None, dict(headers=serialize_headers(headers))
        )

    async def set_offline(self, offline: bool) -> None:
        await self._channel.send("setOffline", None, dict(offline=offline))

    async def add_init_script(
        self, script: str = None, path: Union[str, Path] = None
    ) -> None:
        if path:
            script = (await async_readfile(path)).decode()
        if not isinstance(script, str):
            raise Error("Either path or script parameter must be specified")
        await self._channel.send("addInitScript", None, dict(source=script))

    async def expose_binding(
        self, name: str, callback: Callable, handle: bool = None
    ) -> None:
        for page in self._pages:
            if name in page._bindings:
                raise Error(
                    f'Function "{name}" has been already registered in one of the pages'
                )
        if name in self._bindings:
            raise Error(f'Function "{name}" has been already registered')
        self._bindings[name] = callback
        await self._channel.send(
            "exposeBinding", None, dict(name=name, needsHandle=handle or False)
        )

    async def expose_function(self, name: str, callback: Callable) -> None:
        await self.expose_binding(name, lambda source, *args: callback(*args))

    async def route(
        self, url: URLMatch, handler: RouteHandlerCallback, times: int = None
    ) -> None:
        self._routes.insert(
            0,
            RouteHandler(
                self._base_url,
                url,
                handler,
                True if self._dispatcher_fiber else False,
                times,
            ),
        )
        await self._update_interception_patterns()

    async def unroute(
        self, url: URLMatch, handler: Optional[RouteHandlerCallback] = None
    ) -> None:
        removed = []
        remaining = []
        for route in self._routes:
            if route.url != url or (handler and route.handler != handler):
                remaining.append(route)
            else:
                removed.append(route)
        await self._unroute_internal(removed, remaining, "default")

    async def _unroute_internal(
        self,
        removed: List[RouteHandler],
        remaining: List[RouteHandler],
        behavior: Literal["default", "ignoreErrors", "wait"] = None,
    ) -> None:
        self._routes = remaining
        if behavior is not None and behavior != "default":
            await asyncio.gather(*map(lambda router: router.stop(behavior), removed))  # type: ignore
        await self._update_interception_patterns()

    async def route_web_socket(
        self, url: URLMatch, handler: WebSocketRouteHandlerCallback
    ) -> None:
        self._web_socket_routes.insert(
            0,
            WebSocketRouteHandler(self._base_url, url, handler),
        )
        await self._update_web_socket_interception_patterns()

    def _dispose_har_routers(self) -> None:
        for router in self._har_routers:
            router.dispose()
        self._har_routers = []

    async def unroute_all(
        self, behavior: Literal["default", "ignoreErrors", "wait"] = None
    ) -> None:
        await self._unroute_internal(self._routes, [], behavior)
        self._dispose_har_routers()

    async def _record_into_har(
        self,
        har: Union[Path, str],
        page: Optional[Page] = None,
        url: Union[Pattern[str], str] = None,
        update_content: HarContentPolicy = None,
        update_mode: HarMode = None,
    ) -> None:
        update_content = update_content or "attach"
        params: Dict[str, Any] = {
            "options": {
                "zip": str(har).endswith(".zip"),
                "content": update_content,
                "urlGlob": url if isinstance(url, str) else None,
                "urlRegexSource": url.pattern if isinstance(url, Pattern) else None,
                "urlRegexFlags": (
                    escape_regex_flags(url) if isinstance(url, Pattern) else None
                ),
                "mode": update_mode or "minimal",
            }
        }
        if page:
            params["page"] = page._channel
        har_id = await self._channel.send("harStart", None, params)
        self._har_recorders[har_id] = {
            "path": str(har),
            "content": update_content,
        }

    async def route_from_har(
        self,
        har: Union[Path, str],
        url: Union[Pattern[str], str] = None,
        notFound: RouteFromHarNotFoundPolicy = None,
        update: bool = None,
        updateContent: Literal["attach", "embed"] = None,
        updateMode: HarMode = None,
    ) -> None:
        if update:
            await self._record_into_har(
                har=har,
                page=None,
                url=url,
                update_content=updateContent,
                update_mode=updateMode,
            )
            return
        router = await HarRouter.create(
            local_utils=self._connection.local_utils,
            file=str(har),
            not_found_action=notFound or "abort",
            url_matcher=url,
        )
        self._har_routers.append(router)
        await router.add_context_route(self)

    async def _update_interception_patterns(self) -> None:
        patterns = RouteHandler.prepare_interception_patterns(self._routes)
        await self._channel.send(
            "setNetworkInterceptionPatterns", None, {"patterns": patterns}
        )

    async def _update_web_socket_interception_patterns(self) -> None:
        patterns = WebSocketRouteHandler.prepare_interception_patterns(
            self._web_socket_routes
        )
        await self._channel.send(
            "setWebSocketInterceptionPatterns", None, {"patterns": patterns}
        )

    def expect_event(
        self,
        event: str,
        predicate: Callable = None,
        timeout: float = None,
    ) -> EventContextManagerImpl:
        if timeout is None:
            timeout = self._timeout_settings.timeout()
        waiter = Waiter(self, f"browser_context.expect_event({event})")
        waiter.reject_on_timeout(
            timeout, f'Timeout {timeout}ms exceeded while waiting for event "{event}"'
        )
        if event != BrowserContext.Events.Close:
            waiter.reject_on_event(
                self, BrowserContext.Events.Close, lambda: TargetClosedError()
            )
        waiter.wait_for_event(self, event, predicate)
        return EventContextManagerImpl(waiter.result())

    def _on_close(self) -> None:
        self._closing_or_closed = True
        if self._browser:
            if self in self._browser._contexts:
                self._browser._contexts.remove(self)
            assert self._browser._browser_type is not None
            if (
                self
                in self._browser._browser_type._playwright.selectors._contexts_for_selectors
            ):
                self._browser._browser_type._playwright.selectors._contexts_for_selectors.remove(
                    self
                )

        self._dispose_har_routers()
        self._tracing._reset_stack_counter()
        self.emit(BrowserContext.Events.Close, self)

    async def close(self, reason: str = None) -> None:
        if self._closing_or_closed:
            return
        self._close_reason = reason
        self._closing_or_closed = True

        await self.request.dispose(reason=reason)

        async def _inner_close() -> None:
            for har_id, params in self._har_recorders.items():
                har = cast(
                    Artifact,
                    from_channel(
                        await self._channel.send("harExport", None, {"harId": har_id})
                    ),
                )
                # Server side will compress artifact if content is attach or if file is .zip.
                is_compressed = params.get("content") == "attach" or params[
                    "path"
                ].endswith(".zip")
                need_compressed = params["path"].endswith(".zip")
                if is_compressed and not need_compressed:
                    tmp_path = params["path"] + ".tmp"
                    await har.save_as(tmp_path)
                    await self._connection.local_utils.har_unzip(
                        zipFile=tmp_path, harFile=params["path"]
                    )
                else:
                    await har.save_as(params["path"])
                await har.delete()

        await self._channel._connection.wrap_api_call(_inner_close, True)
        await self._channel.send("close", None, {"reason": reason})
        await self._closed_future

    async def storage_state(
        self, path: Union[str, Path] = None, indexedDB: bool = None
    ) -> StorageState:
        result = await self._channel.send_return_as_dict(
            "storageState", None, {"indexedDB": indexedDB}
        )
        if path:
            await async_writefile(path, json.dumps(result))
        return result

    def _effective_close_reason(self) -> Optional[str]:
        if self._close_reason:
            return self._close_reason
        if self._browser:
            return self._browser._close_reason
        return None

    async def wait_for_event(
        self, event: str, predicate: Callable = None, timeout: float = None
    ) -> Any:
        async with self.expect_event(event, predicate, timeout) as event_info:
            pass
        return await event_info

    def expect_console_message(
        self,
        predicate: Callable[[ConsoleMessage], bool] = None,
        timeout: float = None,
    ) -> EventContextManagerImpl[ConsoleMessage]:
        return self.expect_event(Page.Events.Console, predicate, timeout)

    def expect_page(
        self,
        predicate: Callable[[Page], bool] = None,
        timeout: float = None,
    ) -> EventContextManagerImpl[Page]:
        return self.expect_event(BrowserContext.Events.Page, predicate, timeout)

    def _on_background_page(self, page: Page) -> None:
        self._background_pages.add(page)
        self.emit(BrowserContext.Events.BackgroundPage, page)

    def _on_service_worker(self, worker: Worker) -> None:
        worker._context = self
        self._service_workers.add(worker)
        self.emit(BrowserContext.Events.ServiceWorker, worker)

    def _on_request_failed(
        self,
        request: Request,
        response_end_timing: float,
        failure_text: Optional[str],
        page: Optional[Page],
    ) -> None:
        request._failure_text = failure_text
        request._set_response_end_timing(response_end_timing)
        self.emit(BrowserContext.Events.RequestFailed, request)
        if page:
            page.emit(Page.Events.RequestFailed, request)

    def _on_request_finished(
        self,
        request: Request,
        response: Optional[Response],
        response_end_timing: float,
        page: Optional[Page],
    ) -> None:
        request._set_response_end_timing(response_end_timing)
        self.emit(BrowserContext.Events.RequestFinished, request)
        if page:
            page.emit(Page.Events.RequestFinished, request)
        if response:
            response._finished_future.set_result(True)

    def _on_console_message(self, event: Dict) -> None:
        message = ConsoleMessage(event, self._loop, self._dispatcher_fiber)
        self.emit(BrowserContext.Events.Console, message)
        page = message.page
        if page:
            page.emit(Page.Events.Console, message)

    def _on_dialog(self, dialog: Dialog) -> None:
        has_listeners = self.emit(BrowserContext.Events.Dialog, dialog)
        page = dialog.page
        if page:
            has_listeners = page.emit(Page.Events.Dialog, dialog) or has_listeners
        if not has_listeners:
            # Although we do similar handling on the server side, we still need this logic
            # on the client side due to a possible race condition between two async calls:
            # a) removing "dialog" listener subscription (client->server)
            # b) actual "dialog" event (server->client)
            if dialog.type == "beforeunload":
                asyncio.create_task(dialog.accept())
            else:
                asyncio.create_task(dialog.dismiss())

    def _on_page_error(self, error: Error, page: Optional[Page]) -> None:
        self.emit(
            BrowserContext.Events.WebError,
            WebError(self._loop, self._dispatcher_fiber, page, error),
        )
        if page:
            page.emit(Page.Events.PageError, error)

    def _on_request(self, request: Request, page: Optional[Page]) -> None:
        self.emit(BrowserContext.Events.Request, request)
        if page:
            page.emit(Page.Events.Request, request)

    def _on_response(self, response: Response, page: Optional[Page]) -> None:
        self.emit(BrowserContext.Events.Response, response)
        if page:
            page.emit(Page.Events.Response, response)

    @property
    def background_pages(self) -> List[Page]:
        return list(self._background_pages)

    @property
    def service_workers(self) -> List[Worker]:
        return list(self._service_workers)

    async def new_cdp_session(self, page: Union[Page, Frame]) -> CDPSession:
        page = to_impl(page)
        params = {}
        if isinstance(page, Page):
            params["page"] = page._channel
        elif isinstance(page, Frame):
            params["frame"] = page._channel
        else:
            raise Error("page: expected Page or Frame")
        return from_channel(await self._channel.send("newCDPSession", None, params))

    @property
    def tracing(self) -> Tracing:
        return self._tracing

    @property
    def request(self) -> "APIRequestContext":
        return self._request

    @property
    def clock(self) -> Clock:
        return self._clock



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from pathlib import Path
from typing import Dict, Union

from playwright._impl._connection import ChannelOwner


class Stream(ChannelOwner):
    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)

    async def save_as(self, path: Union[str, Path]) -> None:
        file = await self._loop.run_in_executor(None, lambda: open(path, "wb"))
        while True:
            binary = await self._channel.send("read", None, {"size": 1024 * 1024})
            if not binary:
                break
            await self._loop.run_in_executor(
                None, lambda: file.write(base64.b64decode(binary))
            )
        await self._loop.run_in_executor(None, lambda: file.close())

    async def read_all(self) -> bytes:
        binary = b""
        while True:
            chunk = await self._channel.send("read", None, {"size": 1024 * 1024})
            if not chunk:
                break
            binary += base64.b64decode(chunk)
        return binary



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import math
import uuid
from asyncio.tasks import Task
from typing import Any, Callable, List, Tuple, Union

from pyee import EventEmitter

from playwright._impl._connection import ChannelOwner
from playwright._impl._errors import Error, TimeoutError


class Waiter:
    def __init__(self, channel_owner: ChannelOwner, event: str) -> None:
        self._result: asyncio.Future = asyncio.Future()
        self._wait_id = uuid.uuid4().hex
        self._loop = channel_owner._loop
        self._pending_tasks: List[Task] = []
        self._channel = channel_owner._channel
        self._registered_listeners: List[Tuple[EventEmitter, str, Callable]] = []
        self._logs: List[str] = []
        self._wait_for_event_info_before(self._wait_id, event)

    def _wait_for_event_info_before(self, wait_id: str, event: str) -> None:
        self._channel.send_no_reply(
            "waitForEventInfo",
            None,
            {
                "info": {
                    "waitId": wait_id,
                    "phase": "before",
                    "event": event,
                }
            },
        )

    def _wait_for_event_info_after(self, wait_id: str, error: Exception = None) -> None:
        self._channel._connection.wrap_api_call_sync(
            lambda: self._channel.send_no_reply(
                "waitForEventInfo",
                None,
                {
                    "info": {
                        "waitId": wait_id,
                        "phase": "after",
                        **({"error": str(error)} if error else {}),
                    },
                },
            ),
            True,
        )

    def reject_on_event(
        self,
        emitter: EventEmitter,
        event: str,
        error: Union[Error, Callable[..., Error]],
        predicate: Callable = None,
    ) -> None:
        def listener(event_data: Any = None) -> None:
            if not predicate or predicate(event_data):
                self._reject(error() if callable(error) else error)

        emitter.on(event, listener)
        self._registered_listeners.append((emitter, event, listener))

    def reject_on_timeout(self, timeout: float, message: str) -> None:
        if timeout == 0:
            return

        async def reject() -> None:
            await asyncio.sleep(timeout / 1000)
            self._reject(TimeoutError(message))

        self._pending_tasks.append(self._loop.create_task(reject()))

    def _cleanup(self) -> None:
        for task in self._pending_tasks:
            if not task.done():
                task.cancel()
        for listener in self._registered_listeners:
            listener[0].remove_listener(listener[1], listener[2])

    def _fulfill(self, result: Any) -> None:
        self._cleanup()
        if not self._result.done():
            self._result.set_result(result)
        self._wait_for_event_info_after(self._wait_id)

    def _reject(self, exception: Exception) -> None:
        self._cleanup()
        if exception:
            base_class = TimeoutError if isinstance(exception, TimeoutError) else Error
            exception = base_class(str(exception) + format_log_recording(self._logs))
        if not self._result.done():
            self._result.set_exception(exception)
        self._wait_for_event_info_after(self._wait_id, exception)

    def wait_for_event(
        self,
        emitter: EventEmitter,
        event: str,
        predicate: Callable = None,
    ) -> None:
        def listener(event_data: Any = None) -> None:
            if not predicate or predicate(event_data):
                self._fulfill(event_data)

        emitter.on(event, listener)
        self._registered_listeners.append((emitter, event, listener))

    def result(self) -> asyncio.Future:
        return self._result

    def log(self, message: str) -> None:
        self._logs.append(message)
        try:
            self._channel._connection.wrap_api_call_sync(
                lambda: self._channel.send_no_reply(
                    "waitForEventInfo",
                    None,
                    {
                        "info": {
                            "waitId": self._wait_id,
                            "phase": "log",
                            "message": message,
                        },
                    },
                ),
                True,
            )
        except Exception:
            pass


def throw_on_timeout(timeout: float, exception: Exception) -> asyncio.Task:
    async def throw() -> None:
        await asyncio.sleep(timeout / 1000)
        raise exception

    return asyncio.create_task(throw())


def format_log_recording(log: List[str]) -> str:
    if not log:
        return ""
    header = " logs "
    header_length = 60
    left_length = math.floor((header_length - len(header)) / 2)
    right_length = header_length - len(header) - left_length
    new_line = "\n"
    return f"{new_line}{'=' * left_length}{header}{'=' * right_length}{new_line}{new_line.join(log)}{new_line}{'=' * header_length}"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Dict, Optional, Union, cast

from playwright._impl._api_structures import TracingGroupLocation
from playwright._impl._artifact import Artifact
from playwright._impl._connection import ChannelOwner, from_nullable_channel
from playwright._impl._helper import locals_to_params


class Tracing(ChannelOwner):
    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)
        self._include_sources: bool = False
        self._stacks_id: Optional[str] = None
        self._is_tracing: bool = False
        self._traces_dir: Optional[str] = None

    async def start(
        self,
        name: str = None,
        title: str = None,
        snapshots: bool = None,
        screenshots: bool = None,
        sources: bool = None,
    ) -> None:
        params = locals_to_params(locals())
        self._include_sources = bool(sources)

        await self._channel.send("tracingStart", None, params)
        trace_name = await self._channel.send(
            "tracingStartChunk", None, {"title": title, "name": name}
        )
        await self._start_collecting_stacks(trace_name)

    async def start_chunk(self, title: str = None, name: str = None) -> None:
        params = locals_to_params(locals())
        trace_name = await self._channel.send("tracingStartChunk", None, params)
        await self._start_collecting_stacks(trace_name)

    async def _start_collecting_stacks(self, trace_name: str) -> None:
        if not self._is_tracing:
            self._is_tracing = True
            self._connection.set_is_tracing(True)
        self._stacks_id = await self._connection.local_utils.tracing_started(
            self._traces_dir, trace_name
        )

    async def stop_chunk(self, path: Union[pathlib.Path, str] = None) -> None:
        await self._do_stop_chunk(path)

    async def stop(self, path: Union[pathlib.Path, str] = None) -> None:
        await self._do_stop_chunk(path)
        await self._channel.send(
            "tracingStop",
            None,
        )

    async def _do_stop_chunk(self, file_path: Union[pathlib.Path, str] = None) -> None:
        self._reset_stack_counter()

        if not file_path:
            # Not interested in any artifacts
            await self._channel.send("tracingStopChunk", None, {"mode": "discard"})
            if self._stacks_id:
                await self._connection.local_utils.trace_discarded(self._stacks_id)
            return

        is_local = not self._connection.is_remote

        if is_local:
            result = await self._channel.send_return_as_dict(
                "tracingStopChunk", None, {"mode": "entries"}
            )
            await self._connection.local_utils.zip(
                {
                    "zipFile": str(file_path),
                    "entries": result["entries"],
                    "stacksId": self._stacks_id,
                    "mode": "write",
                    "includeSources": self._include_sources,
                }
            )
            return

        result = await self._channel.send_return_as_dict(
            "tracingStopChunk",
            None,
            {
                "mode": "archive",
            },
        )

        artifact = cast(
            Optional[Artifact],
            from_nullable_channel(result.get("artifact")),
        )

        # The artifact may be missing if the browser closed while stopping tracing.
        if not artifact:
            if self._stacks_id:
                await self._connection.local_utils.trace_discarded(self._stacks_id)
            return

        # Save trace to the final local file.
        await artifact.save_as(file_path)
        await artifact.delete()

        await self._connection.local_utils.zip(
            {
                "zipFile": str(file_path),
                "entries": [],
                "stacksId": self._stacks_id,
                "mode": "append",
                "includeSources": self._include_sources,
            }
        )

    def _reset_stack_counter(self) -> None:
        if self._is_tracing:
            self._is_tracing = False
            self._connection.set_is_tracing(False)

    async def group(self, name: str, location: TracingGroupLocation = None) -> None:
        await self._channel.send("tracingGroup", None, locals_to_params(locals()))

    async def group_end(self) -> None:
        await self._channel.send(
            "tracingGroupEnd",
            None,
        )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from playwright._impl._artifact import Artifact

if TYPE_CHECKING:  # pragma: no cover
    from playwright._impl._page import Page


class Download:
    def __init__(
        self, page: "Page", url: str, suggested_filename: str, artifact: Artifact
    ) -> None:
        self._page = page
        self._loop = page._loop
        self._dispatcher_fiber = page._dispatcher_fiber
        self._url = url
        self._suggested_filename = suggested_filename
        self._artifact = artifact

    def __repr__(self) -> str:
        return f"<Download url={self.url!r} suggested_filename={self.suggested_filename!r}>"

    @property
    def page(self) -> "Page":
        return self._page

    @property
    def url(self) -> str:
        return self._url

    @property
    def suggested_filename(self) -> str:
        return self._suggested_filename

    async def delete(self) -> None:
        await self._artifact.delete()

    async def failure(self) -> Optional[str]:
        return await self._artifact.failure()

    async def path(self) -> pathlib.Path:
        return await self._artifact.path_after_finished()

    async def save_as(self, path: Union[str, Path]) -> None:
        await self._artifact.save_as(path)

    async def cancel(self) -> None:
        return await self._artifact.cancel()



# -*- coding: utf-8 -*-
"""Tests for Beautiful Soup's tree traversal methods.

The tree traversal methods are the main advantage of using Beautiful
Soup over just using a parser.

Different parsers will build different Beautiful Soup trees given the
same markup, but all Beautiful Soup trees can be traversed with the
methods tested here.
"""

import pytest
import re
import warnings
from bs4 import BeautifulSoup
from bs4.builder import builder_registry
from bs4.element import (
    AttributeResemblesVariableWarning,
    CData,
    Comment,
    NavigableString,
    Tag,
)
from bs4.filter import SoupStrainer
from . import (
    SoupTest,
)


class TestFind(SoupTest):
    """Basic tests of the find() method.
    """

    def test_find_tag(self):
        soup = self.soup("<a>1</a><b>2</b><a>3</a><b>4</b>")
        assert soup.find("b").string == "2"

    def test_unicode_text_find(self):
        soup = self.soup("<h1>Räksmörgås</h1>")
        assert soup.find(string="Räksmörgås") == "Räksmörgås"

    def test_unicode_attribute_find(self):
        soup = self.soup('<h1 id="Räksmörgås">here it is</h1>')
        str(soup)
        assert "here it is" == soup.find(id="Räksmörgås").text

    def test_find_everything(self):
        """Test an optimization that finds all tags."""
        soup = self.soup("<a>foo</a><b>bar</b>")
        assert 2 == len(soup.find_all())

    def test_find_everything_with_name(self):
        """Test an optimization that finds all tags with a given name."""
        soup = self.soup("<a>foo</a><b>bar</b><a>baz</a>")
        assert 2 == len(soup.find_all("a"))

    def test_find_with_no_arguments(self):
        soup = self.soup("<div></div><p></p>")
        assert "div" == soup.find().name
        assert "div" == soup.find("p").find_previous_sibling().name
        assert "p" == soup.find("div").find_next_sibling().name

    def test_find_with_no_arguments_only_finds_tags(self):
        soup = self.soup("text<div>text</div>text<p>text</p>")
        assert "div" == soup.find().name
        assert "div" == soup.find("p").find_previous_sibling().name
        assert "p" == soup.find("div").find_next_sibling().name

    def test_find_with_function_can_only_find_tags(self):
        soup = self.soup("text<div>text</div>text<p>text</p>")
        assert "p" == soup.find(lambda t: t.name=="p").name
        assert None == soup.find(lambda t: t=="text")

class TestFindAll(SoupTest):
    """Basic tests of the find_all() method."""

    def test_find_all_with_no_arguments_only_finds_tags(self):
        soup = self.soup("<body>text<div>text</div>text<p>text</p></body>")
        assert 2 == len(soup.body.find_all())
        assert 1 == len(soup.find("p").find_previous_siblings())
        assert 1 == len(soup.find("div").find_next_siblings())

    def test_find_all_text_nodes(self):
        """You can search the tree for text nodes."""
        soup = self.soup("<html>Foo<b>bar</b>\xbb</html>")
        # Exact match.
        assert soup.find_all(string="bar") == ["bar"]

        # Match any of a number of strings.
        assert soup.find_all(string=["Foo", "bar"]) == ["Foo", "bar"]
        # Match a regular expression.
        assert soup.find_all(string=re.compile(".*")) == ["Foo", "bar", "\xbb"]
        # Match anything.
        assert soup.find_all(string=True) == ["Foo", "bar", "\xbb"]

    def test_find_all_limit(self):
        """You can limit the number of items returned by find_all."""
        soup = self.soup("<a>1</a><a>2</a><a>3</a><a>4</a><a>5</a>")
        self.assert_selects(soup.find_all("a", limit=3), ["1", "2", "3"])
        self.assert_selects(soup.find_all("a", limit=1), ["1"])
        self.assert_selects(soup.find_all("a", limit=10), ["1", "2", "3", "4", "5"])

        # A limit of 0 means no limit.
        self.assert_selects(soup.find_all("a", limit=0), ["1", "2", "3", "4", "5"])

    def test_calling_a_tag_is_calling_findall(self):
        soup = self.soup("<a>1</a><b>2<a id='foo'>3</a></b>")
        self.assert_selects(soup("a", limit=1), ["1"])
        self.assert_selects(soup.b(id="foo"), ["3"])

    def test_find_all_with_self_referential_data_structure_does_not_cause_infinite_recursion(
        self,
    ):
        soup = self.soup("<a></a>")
        # Create a self-referential list.
        selfref = []
        selfref.append(selfref)

        # Without special code in SoupStrainer, this would cause infinite
        # recursion.
        with warnings.catch_warnings(record=True) as w:
            assert [] == soup.find_all(selfref)
            [warning] = w
            assert warning.filename == __file__
            msg = str(warning.message)
            assert (
                msg
                == "Ignoring nested list [[...]] to avoid the possibility of infinite recursion."
            )

    def test_find_all_resultset(self):
        """All find_all calls return a ResultSet"""
        soup = self.soup("<a></a>")
        result = soup.find_all("a")
        assert hasattr(result, "source")

        result = soup.find_all(True)
        assert hasattr(result, "source")

        result = soup.find_all(string="foo")
        assert hasattr(result, "source")


class TestFindAllBasicNamespaces(SoupTest):
    def test_find_by_namespaced_name(self):
        soup = self.soup('<mathml:msqrt>4</mathml:msqrt><a svg:fill="red">')
        assert "4" == soup.find("mathml:msqrt").string
        assert "a" == soup.find(attrs={"svg:fill": "red"}).name


class TestFindAllByName(SoupTest):
    """Test ways of finding tags by tag name."""

    def setup_method(self) -> None:
        self.tree = self.soup("""<a>First tag.</a>
                                  <b>Second tag.</b>
                                  <c>Third <a>Nested tag.</a> tag.</c>""")

    def test_find_all_by_tag_name(self):
        # Find all the <a> tags.
        self.assert_selects(self.tree.find_all("a"), ["First tag.", "Nested tag."])

    def test_find_all_by_name_and_text(self):
        self.assert_selects(
            self.tree.find_all("a", string="First tag."), ["First tag."]
        )

        self.assert_selects(
            self.tree.find_all("a", string=True), ["First tag.", "Nested tag."]
        )

        self.assert_selects(
            self.tree.find_all("a", string=re.compile("tag")),
            ["First tag.", "Nested tag."],
        )

    def test_find_all_on_non_root_element(self):
        # You can call find_all on any node, not just the root.
        self.assert_selects(self.tree.c.find_all("a"), ["Nested tag."])

    def test_calling_element_invokes_find_all(self):
        self.assert_selects(self.tree("a"), ["First tag.", "Nested tag."])

    def test_find_all_by_tag_strainer(self):
        self.assert_selects(
            self.tree.find_all(SoupStrainer("a")), ["First tag.", "Nested tag."]
        )

    def test_find_all_by_tag_names(self):
        self.assert_selects(
            self.tree.find_all(["a", "b"]), ["First tag.", "Second tag.", "Nested tag."]
        )

    def test_find_all_by_tag_dict(self):
        self.assert_selects(
            self.tree.find_all({"a": True, "b": True}),
            ["First tag.", "Second tag.", "Nested tag."],
        )

    def test_find_all_by_tag_re(self):
        self.assert_selects(
            self.tree.find_all(re.compile("^[ab]$")),
            ["First tag.", "Second tag.", "Nested tag."],
        )

    def test_find_all_with_tags_matching_method(self):
        # You can define an oracle method that determines whether
        # a tag matches the search.
        def id_matches_name(tag):
            return tag.name == tag.get("id")

        tree = self.soup("""<a id="a">Match 1.</a>
                            <a id="1">Does not match.</a>
                            <b id="b">Match 2.</a>""")

        self.assert_selects(tree.find_all(id_matches_name), ["Match 1.", "Match 2."])

    def test_find_with_multi_valued_attribute(self):
        soup = self.soup(
            "<div class='a b'>1</div><div class='a c'>2</div><div class='a d'>3</div>"
        )
        r1 = soup.find("div", "a d")
        r2 = soup.find("div", re.compile(r"a d"))
        r3, r4 = soup.find_all("div", ["a b", "a d"])
        assert "3" == r1.string
        assert "3" == r2.string
        assert "1" == r3.string
        assert "3" == r4.string


class TestFindAllByAttribute(SoupTest):
    def test_find_all_by_attribute_name(self):
        # You can pass in keyword arguments to find_all to search by
        # attribute.
        tree = self.soup("""
                         <a id="first">Matching a.</a>
                         <a id="second">
                          Non-matching <b id="first">Matching b.</b>a.
                         </a>""")
        self.assert_selects(tree.find_all(id="first"), ["Matching a.", "Matching b."])

    def test_find_all_by_utf8_attribute_value(self):
        peace = "םולש".encode("utf8")
        data = '<a title="םולש"></a>'.encode("utf8")
        soup = self.soup(data)
        assert [soup.a] == soup.find_all(title=peace)
        assert [soup.a] == soup.find_all(title=peace.decode("utf8"))
        assert [soup.a], soup.find_all(title=[peace, "something else"])

    def test_find_all_by_attribute_dict(self):
        # You can pass in a dictionary as the argument 'attrs'. This
        # lets you search for attributes like 'name' (a fixed argument
        # to find_all) and 'class' (a reserved word in Python.)
        tree = self.soup("""
                         <a name="name1" class="class1">Name match.</a>
                         <a name="name2" class="class2">Class match.</a>
                         <a name="name3" class="class3">Non-match.</a>
                         <name1>A tag called 'name1'.</name1>
                         """)

        # This doesn't do what you want.
        self.assert_selects(tree.find_all(name="name1"), ["A tag called 'name1'."])
        # This does what you want.
        self.assert_selects(tree.find_all(attrs={"name": "name1"}), ["Name match."])

        self.assert_selects(tree.find_all(attrs={"class": "class2"}), ["Class match."])

    def test_find_all_by_class(self):
        tree = self.soup("""
                         <a class="1">Class 1.</a>
                         <a class="2">Class 2.</a>
                         <b class="1">Class 1.</b>
                         <c class="3 4">Class 3 and 4.</c>
                         """)

        # Passing in the class_ keyword argument will search against
        # the 'class' attribute.
        self.assert_selects(tree.find_all("a", class_="1"), ["Class 1."])
        self.assert_selects(tree.find_all("c", class_="3"), ["Class 3 and 4."])
        self.assert_selects(tree.find_all("c", class_="4"), ["Class 3 and 4."])

        # Passing in a string to 'attrs' will also search the CSS class.
        self.assert_selects(tree.find_all("a", "1"), ["Class 1."])
        self.assert_selects(tree.find_all(attrs="1"), ["Class 1.", "Class 1."])
        self.assert_selects(tree.find_all("c", "3"), ["Class 3 and 4."])
        self.assert_selects(tree.find_all("c", "4"), ["Class 3 and 4."])

    def test_find_by_class_when_multiple_classes_present(self):
        tree = self.soup("<gar class='foo bar'>Found it</gar>")

        f = tree.find_all("gar", class_=re.compile("o"))
        self.assert_selects(f, ["Found it"])

        f = tree.find_all("gar", class_=re.compile("a"))
        self.assert_selects(f, ["Found it"])

        # If the search fails to match the individual strings "foo" and "bar",
        # it will be tried against the combined string "foo bar".
        f = tree.find_all("gar", class_=re.compile("o b"))
        self.assert_selects(f, ["Found it"])

    def test_find_all_with_non_dictionary_for_attrs_finds_by_class(self):
        soup = self.soup("<a class='bar'>Found it</a>")

        self.assert_selects(soup.find_all("a", re.compile("ba")), ["Found it"])

        def big_attribute_value(value):
            return len(value) > 3

        self.assert_selects(soup.find_all("a", big_attribute_value), [])

        def small_attribute_value(value):
            return len(value) <= 3

        self.assert_selects(soup.find_all("a", small_attribute_value), ["Found it"])

    def test_find_all_with_string_for_attrs_finds_multiple_classes(self):
        soup = self.soup('<a class="foo bar"></a><a class="foo"></a>')
        a, a2 = soup.find_all("a")
        assert [a, a2], soup.find_all("a", "foo")
        assert [a], soup.find_all("a", "bar")

        # If you specify the class as a string that contains a
        # space, only that specific value will be found.
        assert [a] == soup.find_all("a", class_="foo bar")
        assert [a] == soup.find_all("a", "foo bar")
        assert [] == soup.find_all("a", "bar foo")

    def test_find_all_by_attribute_soupstrainer(self):
        tree = self.soup("""
                         <a id="first">Match.</a>
                         <a id="second">Non-match.</a>""")

        strainer = SoupStrainer(attrs={"id": "first"})
        self.assert_selects(tree.find_all(strainer), ["Match."])

    def test_find_all_with_missing_attribute(self):
        # You can pass in None as the value of an attribute to find_all.
        # This will match tags that do not have that attribute set.
        tree = self.soup("""<a id="1">ID present.</a>
                            <a>No ID present.</a>
                            <a id="">ID is empty.</a>""")
        self.assert_selects(tree.find_all("a", id=None), ["No ID present."])

    def test_find_all_with_defined_attribute(self):
        # You can pass in None as the value of an attribute to find_all.
        # This will match tags that have that attribute set to any value.
        tree = self.soup("""<a id="1">ID present.</a>
                            <a>No ID present.</a>
                            <a id="">ID is empty.</a>""")
        self.assert_selects(tree.find_all(id=True), ["ID present.", "ID is empty."])

    def test_find_all_with_numeric_attribute(self):
        # If you search for a number, it's treated as a string.
        tree = self.soup("""<a id=1>Unquoted attribute.</a>
                            <a id="1">Quoted attribute.</a>""")

        expected = ["Unquoted attribute.", "Quoted attribute."]
        self.assert_selects(tree.find_all(id=1), expected)
        self.assert_selects(tree.find_all(id="1"), expected)

    def test_find_all_with_list_attribute_values(self):
        # You can pass a list of attribute values instead of just one,
        # and you'll get tags that match any of the values.
        tree = self.soup("""<a id="1">1</a>
                            <a id="2">2</a>
                            <a id="3">3</a>
                            <a>No ID.</a>""")
        self.assert_selects(tree.find_all(id=["1", "3", "4"]), ["1", "3"])

        # If you pass in an empty list, you get nothing.
        self.assert_selects(tree.find_all(id=[]), [])

    def test_find_all_with_regular_expression_attribute_value(self):
        # You can pass a regular expression as an attribute value, and
        # you'll get tags whose values for that attribute match the
        # regular expression.
        tree = self.soup("""<a id="a">One a.</a>
                            <a id="aa">Two as.</a>
                            <a id="ab">Mixed as and bs.</a>
                            <a id="b">One b.</a>
                            <a>No ID.</a>""")

        self.assert_selects(tree.find_all(id=re.compile("^a+$")), ["One a.", "Two as."])

    def test_find_by_name_and_containing_string(self):
        soup = self.soup("<b>foo</b><b>bar</b><a>foo</a>")
        a = soup.a

        assert [a] == soup.find_all("a", string="foo")
        assert [] == soup.find_all("a", string="bar")

    def test_find_by_name_and_containing_string_when_string_is_buried(self):
        soup = self.soup("<a>foo</a><a><b><c>foo</c></b></a>")
        assert soup.find_all("a") == soup.find_all("a", string="foo")

    def test_find_by_attribute_and_containing_string(self):
        soup = self.soup('<b id="1">foo</b><a id="2">foo</a>')
        a = soup.a

        assert [a] == soup.find_all(id=2, string="foo")
        assert [] == soup.find_all(id=1, string="bar")


class TestSmooth(SoupTest):
    """Test Tag.smooth."""

    def test_smooth(self):
        soup = self.soup("<div>a</div>")
        div = soup.div
        div.append("b")
        div.append("c")
        div.append(Comment("Comment 1"))
        div.append(Comment("Comment 2"))
        div.append("d")
        builder = self.default_builder()
        span = Tag(soup, builder, "span")
        span.append("1")
        span.append("2")
        div.append(span)

        # At this point the tree has a bunch of adjacent
        # NavigableStrings. This is normal, but it has no meaning in
        # terms of HTML, so we may want to smooth things out for
        # output.

        # Since the <span> tag has two children, its .string is None.
        assert None is div.span.string

        assert 7 == len(div.contents)
        div.smooth()
        assert 5 == len(div.contents)

        # The three strings at the beginning of div.contents have been
        # merged into on string.
        #
        assert "abc" == div.contents[0]

        # The call is recursive -- the <span> tag was also smoothed.
        assert "12" == div.span.string

        # The two comments have _not_ been merged, even though
        # comments are strings. Merging comments would change the
        # meaning of the HTML.
        assert "Comment 1" == div.contents[1]
        assert "Comment 2" == div.contents[2]


class TestIndex(SoupTest):
    """Test Tag.index"""

    def test_index(self):
        tree = self.soup("""<div>
                            <a>Identical</a>
                            <b>Not identical</b>
                            <a>Identical</a>

                            <c><d>Identical with child</d></c>
                            <b>Also not identical</b>
                            <c><d>Identical with child</d></c>
                            </div>""")
        div = tree.div
        for i, element in enumerate(div.contents):
            assert i == div.index(element)
        with pytest.raises(ValueError):
            tree.index(1)


class TestParentOperations(SoupTest):
    """Test navigation and searching through an element's parents."""

    def setup_method(self) -> None:
        self.tree = self.soup("""<ul id="empty"></ul>
                                 <ul id="top">
                                  <ul id="middle">
                                   <ul id="bottom">
                                    <b id="start">Start here</b>
                                   </ul>
                                  </ul>""")
        self.start = self.tree.b

    def test_parent(self):
        assert self.start.parent["id"] == "bottom"
        assert self.start.parent.parent["id"] == "middle"
        assert self.start.parent.parent.parent["id"] == "top"

    def test_parent_of_top_tag_is_soup_object(self):
        top_tag = self.tree.contents[0]
        assert top_tag.parent == self.tree

    def test_soup_object_has_no_parent(self):
        assert None is self.tree.parent

    def test_find_parents(self):
        self.assert_selects_ids(
            self.start.find_parents("ul"), ["bottom", "middle", "top"]
        )
        self.assert_selects_ids(self.start.find_parents("ul", id="middle"), ["middle"])
        assert self.start.find_parents(id="start") == []

    def test_find_parent(self):
        # assert self.start.find_parent('ul')['id'] == 'bottom'
        assert self.start.find_parent("ul", id="top")["id"] == "top"

        assert self.start.find_parent(id="start") is None

    def test_parent_of_text_element(self):
        text = self.tree.find(string="Start here")
        assert text.parent.name == "b"

    def test_text_element_find_parent(self):
        text = self.tree.find(string="Start here")
        assert text.find_parent("ul")["id"] == "bottom"

    def test_parent_generator(self):
        parents = [
            parent["id"]
            for parent in self.start.parents
            if parent is not None and "id" in parent.attrs
        ]
        assert parents == ["bottom", "middle", "top"]

    def test_self_and_parent_generator(self):
        results = [
            parent["id"]
            for parent in self.start.self_and_parents
            if parent is not None and "id" in parent.attrs
        ]
        assert results == ["start", "bottom", "middle", "top"]


class ProximityTest(SoupTest):
    def setup_method(self) -> None:
        self.tree = self.soup(
            '<html id="start"><head id="headtag"></head><body id="bodytag"><b id="1">One</b><b id="2">Two</b><b id="3">Three</b></body></html>'
        )


class TestNextOperations(ProximityTest):
    def setup_method(self) -> None:
        super(TestNextOperations, self).setup_method()
        self.start = self.tree.b

    def test_next(self):
        assert self.start.next_element == "One"
        assert self.start.next_element.next_element["id"] == "2"

    def test_next_of_last_item_is_none(self):
        last = self.tree.find(string="Three")
        assert last.next_element is None

    def test_next_of_root_is_none(self):
        # The document root is outside the next/previous chain.
        assert self.tree.next_element is None

    def test_find_all_next(self):
        self.assert_selects(self.start.find_all_next("b"), ["Two", "Three"])
        self.start.find_all_next(id=3)
        self.assert_selects(self.start.find_all_next(id=3), ["Three"])

    def test_find_next(self):
        assert self.start.find_next("b")["id"] == "2"
        assert self.start.find_next(string="Three") == "Three"

    def test_find_next_for_text_element(self):
        text = self.tree.find(string="One")
        assert text.find_next("b").string == "Two"
        self.assert_selects(text.find_all_next("b"), ["Two", "Three"])

    def test_next_generators(self):
        start = self.tree.find(string="Two")
        successors = [node for node in start.next_elements]
        # There are two successors: the final <b> tag and its text contents.
        tag, contents = successors
        assert tag["id"] == "3"
        assert contents == "Three"

        successors2 = [node for node in start.self_and_next_elements]
        assert successors2[1:] == successors
        assert successors2[0] == start


class TestPreviousOperations(ProximityTest):
    def setup_method(self) -> None:
        super(TestPreviousOperations, self).setup_method()
        self.end = self.tree.find(string="Three")

    def test_previous(self):
        assert self.end.previous_element["id"] == "3"
        assert self.end.previous_element.previous_element == "Two"

    def test_previous_of_first_item_is_none(self):
        first = self.tree.find("html")
        assert first.previous_element is None

    def test_previous_of_root_is_none(self):
        # The document root is outside the next/previous chain.
        assert self.tree.previous_element is None

    def test_find_all_previous(self):
        # The <b> tag containing the "Three" node is the predecessor
        # of the "Three" node itself, which is why "Three" shows up
        # here.
        self.assert_selects(self.end.find_all_previous("b"), ["Three", "Two", "One"])
        self.assert_selects(self.end.find_all_previous(id=1), ["One"])

    def test_find_previous(self):
        assert self.end.find_previous("b")["id"] == "3"
        assert self.end.find_previous(string="One") == "One"

    def test_find_previous_for_text_element(self):
        text = self.tree.find(string="Three")
        assert text.find_previous("b").string == "Three"
        self.assert_selects(text.find_all_previous("b"), ["Three", "Two", "One"])

    def test_previous_generators(self):
        start = self.tree.find("b", string="One")
        self.assert_selects_ids(start.previous_elements, ["bodytag", "headtag", 'start'])
        self.assert_selects_ids(start.self_and_previous_elements, ["1", "bodytag", "headtag", "start"])


class SiblingTest(SoupTest):
    def setup_method(self) -> None:
        markup = """<html>
                    <span id="1">
                     <span id="1.1"></span>
                    </span>
                    <span id="2">
                     <span id="2.1"></span>
                    </span>
                    <span id="3">
                     <span id="3.1"></span>
                    </span>
                    <span id="4"></span>
                    </html>"""
        # All that whitespace looks good but makes the tests more
        # difficult. Get rid of it.
        markup = re.compile(r"\n\s*").sub("", markup)
        self.tree = self.soup(markup)


class TestNextSibling(SiblingTest):
    def setup_method(self) -> None:
        super(TestNextSibling, self).setup_method()
        self.start = self.tree.find(id="1")

    def test_next_sibling_of_root_is_none(self):
        assert self.tree.next_sibling is None

    def test_next_sibling(self):
        assert self.start.next_sibling["id"] == "2"
        assert self.start.next_sibling.next_sibling["id"] == "3"

        # Note the difference between next_sibling and next_element.
        assert self.start.next_element["id"] == "1.1"

    def test_next_sibling_may_not_exist(self):
        assert self.tree.html.next_sibling is None

        nested_span = self.tree.find(id="1.1")
        assert nested_span.next_sibling is None

        last_span = self.tree.find(id="4")
        assert last_span.next_sibling is None

    def test_find_next_sibling(self):
        assert self.start.find_next_sibling("span")["id"] == "2"

    def test_next_siblings(self):
        self.assert_selects_ids(self.start.find_next_siblings("span"), ["2", "3", "4"])

        self.assert_selects_ids(self.start.find_next_siblings(id="3"), ["3"])

    def test_next_siblings_generators(self):
        self.assert_selects_ids(self.start.next_siblings, ["2", "3", "4"])
        self.assert_selects_ids(self.start.self_and_next_siblings, ["1", "2", "3", "4"])

    def test_next_sibling_for_text_element(self):
        soup = self.soup("Foo<b>bar</b>baz")
        start = soup.find(string="Foo")
        assert start.next_sibling.name == "b"
        assert start.next_sibling.next_sibling == "baz"

        self.assert_selects(start.find_next_siblings("b"), ["bar"])
        assert start.find_next_sibling(string="baz") == "baz"
        assert start.find_next_sibling(string="nonesuch") is None


class TestPreviousSibling(SiblingTest):
    def setup_method(self) -> None:
        super(TestPreviousSibling, self).setup_method()
        self.end = self.tree.find(id="4")

    def test_previous_sibling_of_root_is_none(self):
        assert self.tree.previous_sibling is None

    def test_previous_sibling(self):
        assert self.end.previous_sibling["id"] == "3"
        assert self.end.previous_sibling.previous_sibling["id"] == "2"

        # Note the difference between previous_sibling and previous_element.
        assert self.end.previous_element["id"] == "3.1"

    def test_previous_sibling_may_not_exist(self):
        assert self.tree.html.previous_sibling is None

        nested_span = self.tree.find(id="1.1")
        assert nested_span.previous_sibling is None

        first_span = self.tree.find(id="1")
        assert first_span.previous_sibling is None

    def test_find_previous_sibling(self):
        assert self.end.find_previous_sibling("span")["id"] == "3"

    def test_previous_siblings(self):
        self.assert_selects_ids(
            self.end.find_previous_siblings("span"), ["3", "2", "1"]
        )

        self.assert_selects_ids(self.end.find_previous_siblings(id="1"), ["1"])

    def test_previous_siblings_generators(self):
        self.assert_selects_ids(self.end.previous_siblings, ["3", "2", "1"])
        self.assert_selects_ids(self.end.self_and_previous_siblings, ["4", "3", "2", "1"])

    def test_previous_sibling_for_text_element(self):
        soup = self.soup("Foo<b>bar</b>baz")
        start = soup.find(string="baz")
        assert start.previous_sibling.name == "b"
        assert start.previous_sibling.previous_sibling == "Foo"

        self.assert_selects(start.find_previous_siblings("b"), ["bar"])
        assert start.find_previous_sibling(string="Foo") == "Foo"
        assert start.find_previous_sibling(string="nonesuch") is None


class TestTreeModification(SoupTest):
    def test_attribute_modification(self):
        soup = self.soup('<a id="1"></a>')
        soup.a["id"] = 2
        assert soup.decode() == self.document_for('<a id="2"></a>')
        del soup.a["id"]
        assert soup.decode() == self.document_for("<a></a>")
        soup.a["id2"] = "foo"
        assert soup.decode() == self.document_for('<a id2="foo"></a>')

    def test_new_tag_creation(self):
        builder = builder_registry.lookup("html")()
        soup = self.soup("<body></body>", builder=builder)
        a = Tag(soup, builder, "a")
        ol = Tag(soup, builder, "ol")
        a["href"] = "http://foo.com/"
        soup.body.insert(0, a)
        soup.body.insert(1, ol)
        assert (
            soup.body.encode()
            == b'<body><a href="http://foo.com/"></a><ol></ol></body>'
        )

    def test_append_to_contents_moves_tag(self):
        doc = """<p id="1">Don't leave me <b>here</b>.</p>
                <p id="2">Don\'t leave!</p>"""
        soup = self.soup(doc)
        second_para = soup.find(id="2")
        bold = soup.b

        # Move the <b> tag to the end of the second paragraph.
        soup.find(id="2").append(soup.b)

        # The <b> tag is now a child of the second paragraph.
        assert bold.parent == second_para

        assert soup.decode() == self.document_for(
            '<p id="1">Don\'t leave me .</p>\n' '<p id="2">Don\'t leave!<b>here</b></p>'
        )

    def test_insertion_returns_inserted_things(self):
        soup = self.soup("<html></html>")
        html = soup.find('html')
        head = html.append(soup.new_tag('head'))
        assert head.name == 'head'

        [title] = head.insert(0, soup.new_tag('title'))
        assert title.name == 'title'

        text5 = title.append('5')
        assert text5 == '5'
        text34 = text5.insert_before('3', '4')
        assert text34 == ['3', '4']
        text67 = text5.insert_after('6', '7')
        assert text67 == ['6', '7']
        text89 = title.extend(['8', '9'])
        assert text89 == ['8', '9']
        assert title.get_text() == '3456789'

    def test_replace_with_returns_thing_that_was_replaced(self):
        text = "<a></a><b><c></c></b>"
        soup = self.soup(text)
        a = soup.a
        new_a = a.replace_with(soup.c)
        assert a == new_a

    def test_unwrap_returns_thing_that_was_replaced(self):
        text = "<a><b></b><c></c></a>"
        soup = self.soup(text)
        a = soup.a
        new_a = a.unwrap()
        assert a == new_a

    def test_replace_with_and_unwrap_give_useful_exception_when_tag_has_no_parent(self):
        soup = self.soup("<a><b>Foo</b></a><c>Bar</c>")
        a = soup.a
        a.extract()
        assert None is a.parent
        with pytest.raises(ValueError):
            a.unwrap()
        with pytest.raises(ValueError):
            a.replace_with(soup.c)

    def test_replace_tag_with_itself(self):
        text = "<a><b></b><c>Foo<d></d></c></a><a><e></e></a>"
        soup = self.soup(text)
        c = soup.c
        result = soup.c.replace_with(c)
        assert result == c
        assert soup.decode() == self.document_for(text)

    def test_replace_tag_with_its_parent_raises_exception(self):
        text = "<a><b></b></a>"
        soup = self.soup(text)
        with pytest.raises(ValueError):
            soup.b.replace_with(soup.a)

    def test_insert_tag_into_itself_raises_exception(self):
        text = "<a><b></b></a>"
        soup = self.soup(text)
        with pytest.raises(ValueError):
            soup.a.insert(0, soup.a)

    def test_insert_multiple_elements(self):
        soup = self.soup("<p>And now, a word:</p><p>And we're back.</p>")
        p2, p3 = soup.insert(1, soup.new_tag("p", string="p2"), soup.new_tag("p", string="p3"))
        assert "p2" == p2.string
        assert "p3" == p3.string

        p1, p2, p3, p4 = list(soup.children)
        assert "And now, a word:" == p1.string
        assert "p2" == p2.string
        assert "p3" == p3.string
        assert "And we're back." == p4.string

    def test_insert_beautifulsoup_object_inserts_children(self):
        """Inserting one BeautifulSoup object into another actually inserts all
        of its children -- you'll never combine BeautifulSoup objects.
        """
        soup = self.soup("<p>And now, a word:</p><p>And we're back.</p>")

        text = "<p>p2</p><p>p3</p>"
        to_insert = self.soup(text)
        p2, p3 = soup.insert(1, to_insert)
        assert "p2" == p2.string
        assert "p3" == p3.string

        for i in soup.descendants:
            assert not isinstance(i, BeautifulSoup)

        p1, p2, p3, p4 = list(soup.children)
        assert "And now, a word:" == p1.string
        assert "p2" == p2.string
        assert "p3" == p3.string
        assert "And we're back." == p4.string

    def test_replace_with_maintains_next_element_throughout(self):
        soup = self.soup("<p><a>one</a><b>three</b></p>")
        a = soup.a
        # Make it so the <a> tag has two text children.
        a.insert(1, "two")

        # Now replace each one with the empty string.
        left, right = a.contents
        left.replace_with("")
        right.replace_with("")

        # The <b> tag is still connected to the tree.
        assert "three" == soup.b.string

    def test_replace_final_node(self):
        soup = self.soup("<b>Argh!</b>")
        soup.find(string="Argh!").replace_with("Hooray!")
        new_text = soup.find(string="Hooray!")
        b = soup.b
        assert new_text.previous_element == b
        assert new_text.parent == b
        assert new_text.previous_element.next_element == new_text
        assert new_text.next_element is None

    def test_consecutive_text_nodes(self):
        # A builder should never create two consecutive text nodes,
        # but if you insert one next to another, Beautiful Soup will
        # handle it correctly.
        soup = self.soup("<a><b>Argh!</b><c></c></a>")
        soup.b.insert(1, "Hooray!")

        assert soup.decode() == self.document_for("<a><b>Argh!Hooray!</b><c></c></a>")

        new_text = soup.find(string="Hooray!")
        assert new_text.previous_element == "Argh!"
        assert new_text.previous_element.next_element == new_text

        assert new_text.previous_sibling == "Argh!"
        assert new_text.previous_sibling.next_sibling == new_text

        assert new_text.next_sibling is None
        assert new_text.next_element == soup.c

    def test_insert_string(self):
        soup = self.soup("<a></a>")
        soup.a.insert(0, "bar")
        soup.a.insert(0, "foo")
        # The string were added to the tag.
        assert ["foo", "bar"] == soup.a.contents
        # And they were converted to NavigableStrings.
        assert soup.a.contents[0].next_element == "bar"

    def test_append(self):
        soup = self.soup("<b>1</b>")
        result = soup.b.append("2")
        assert result == "2"
        assert soup.b.decode() == "<b>12</b>"

    def test_insert_tag(self):
        builder = self.default_builder()
        soup = self.soup("<a><b>Find</b><c>lady!</c><d></d></a>", builder=builder)
        magic_tag = Tag(soup, builder, "magictag")
        magic_tag.insert(0, "the")
        soup.a.insert(1, magic_tag)

        assert soup.decode() == self.document_for(
            "<a><b>Find</b><magictag>the</magictag><c>lady!</c><d></d></a>"
        )

        # Make sure all the relationships are hooked up correctly.
        b_tag = soup.b
        assert b_tag.next_sibling == magic_tag
        assert magic_tag.previous_sibling == b_tag

        find = b_tag.find(string="Find")
        assert find.next_element == magic_tag
        assert magic_tag.previous_element == find

        c_tag = soup.c
        assert magic_tag.next_sibling == c_tag
        assert c_tag.previous_sibling == magic_tag

        the = magic_tag.find(string="the")
        assert the.parent == magic_tag
        assert the.next_element == c_tag
        assert c_tag.previous_element == the

    def test_insert_into_the_current_location(self):
        data = "<a>b<c></c>d</a>"
        soup = self.soup(data)
        soup.a.insert(1, soup.c)
        assert data == soup.decode()

    def test_append_child_thats_already_at_the_end(self):
        data = "<a><b></b></a>"
        soup = self.soup(data)
        soup.a.append(soup.b)
        assert data == soup.decode()

    def test_extend_with_a_list_of_elements(self):
        data = "<a><b><c><d><e><f><g></g></f></e></d></c></b></a>"
        soup = self.soup(data)
        elements = [soup.g, soup.f, soup.e, soup.d, soup.c, soup.b]
        soup.a.extend(elements)
        assert "<a><g></g><f></f><e></e><d></d><c></c><b></b></a>" == soup.decode()

    def test_extend_with_a_list_of_strings(self):
        data = "<a></a>"
        soup = self.soup(data)
        elements = ["b", "c", NavigableString("d"), "e"]
        soup.a.extend(elements)
        assert "<a>bcde</a>" == soup.decode()

    @pytest.mark.parametrize("get_tags", [lambda tag: tag, lambda tag: tag.contents])
    def test_extend_with_another_tags_contents(self, get_tags):
        data = '<body><div id="d1"><a>1</a><a>2</a><a>3</a><a>4</a></div><div id="d2"></div></body>'
        soup = self.soup(data)
        d1 = soup.find("div", id="d1")
        d2 = soup.find("div", id="d2")
        tags = get_tags(d1)
        d2.extend(tags)
        assert '<div id="d1"></div>' == d1.decode()
        assert '<div id="d2"><a>1</a><a>2</a><a>3</a><a>4</a></div>' == d2.decode()

    @pytest.mark.parametrize(
        "string_source,result",
        (
            [lambda soup: soup.a.string, "<a></a><b>1</b>"],
            [lambda soup: "abcde", "<a>1</a><b>abcde</b>"],
        ),
    )
    def test_extend_with_a_single_non_tag_element(self, string_source, result):
        data = "<div><a>1</a><b></b></div>"
        soup = self.soup(data)
        with warnings.catch_warnings(record=True) as w:
            string = string_source(soup)
            soup.b.extend(string)
            assert soup.div.decode_contents() == result
            [warning] = w
            assert warning.filename == __file__
            msg = str(warning.message)
            assert (
                msg
                == "A single non-Tag item was passed into Tag.extend. Use Tag.append instead."
            )

    def test_move_tag_to_beginning_of_parent(self):
        data = "<a><b></b><c></c><d></d></a>"
        soup = self.soup(data)
        soup.a.insert(0, soup.d)
        assert "<a><d></d><b></b><c></c></a>" == soup.decode()

    def test_insert_works_on_empty_element_tag(self):
        # This is a little strange, since most HTML parsers don't allow
        # markup like this to come through. But in general, we don't
        # know what the parser would or wouldn't have allowed, so
        # I'm letting this succeed for now.
        soup = self.soup("<br/>")
        soup.br.insert(1, "Contents")
        assert str(soup.br) == "<br>Contents</br>"

    def test_insert_before(self):
        soup = self.soup("<a>foo</a><b>bar</b>")
        soup.b.insert_before("BAZ")
        soup.a.insert_before("QUUX")
        assert soup.decode() == self.document_for("QUUX<a>foo</a>BAZ<b>bar</b>")

        soup.a.insert_before(soup.b)
        assert soup.decode() == self.document_for("QUUX<b>bar</b><a>foo</a>BAZ")

        # Can't insert an element before itself.
        b = soup.b
        with pytest.raises(ValueError):
            b.insert_before(b)

        # Can't insert before if an element has no parent.
        b.extract()
        with pytest.raises(ValueError):
            b.insert_before("nope")

        # Can insert an identical element
        soup = self.soup("<a>")
        soup.a.insert_before(soup.new_tag("a"))

        # TODO: OK but what happens?

    def test_insert_multiple_before(self):
        soup = self.soup("<a>foo</a><b>bar</b>")
        soup.b.insert_before("BAZ", " ", "QUUX")
        soup.a.insert_before("QUUX", " ", "BAZ")
        assert soup.decode() == self.document_for(
            "QUUX BAZ<a>foo</a>BAZ QUUX<b>bar</b>"
        )

        soup.a.insert_before(soup.b, "FOO")
        assert soup.decode() == self.document_for(
            "QUUX BAZ<b>bar</b>FOO<a>foo</a>BAZ QUUX"
        )

    def test_insert_after(self):
        soup = self.soup("<a>foo</a><b>bar</b>")
        soup.b.insert_after("BAZ")
        soup.a.insert_after("QUUX")
        assert soup.decode() == self.document_for("<a>foo</a>QUUX<b>bar</b>BAZ")
        soup.b.insert_after(soup.a)
        assert soup.decode() == self.document_for("QUUX<b>bar</b><a>foo</a>BAZ")

        # Can't insert an element after itself.
        b = soup.b
        with pytest.raises(ValueError):
            b.insert_after(b)

        # Can't insert after if an element has no parent.
        b.extract()
        with pytest.raises(ValueError):
            b.insert_after("nope")

        # Can insert an identical element
        soup = self.soup("<a>")
        soup.a.insert_before(soup.new_tag("a"))

        # TODO: OK but what does it look like?

    def test_insert_multiple_after(self):
        soup = self.soup("<a>foo</a><b>bar</b>")
        soup.b.insert_after("BAZ", " ", "QUUX")
        soup.a.insert_after("QUUX", " ", "BAZ")
        assert soup.decode() == self.document_for(
            "<a>foo</a>QUUX BAZ<b>bar</b>BAZ QUUX"
        )
        soup.b.insert_after(soup.a, "FOO ")
        assert soup.decode() == self.document_for(
            "QUUX BAZ<b>bar</b><a>foo</a>FOO BAZ QUUX"
        )

    def test_insert_after_raises_exception_if_after_has_no_meaning(self):
        soup = self.soup("")
        tag = soup.new_tag("a")
        string = soup.new_string("")
        with pytest.raises(ValueError):
            string.insert_after(tag)
        with pytest.raises(NotImplementedError):
            soup.insert_after(tag)
        with pytest.raises(ValueError):
            tag.insert_after(tag)

    def test_insert_before_raises_notimplementederror_if_before_has_no_meaning(self):
        soup = self.soup("")
        tag = soup.new_tag("a")
        string = soup.new_string("")
        with pytest.raises(ValueError):
            string.insert_before(tag)
        with pytest.raises(NotImplementedError):
            soup.insert_before(tag)
        with pytest.raises(ValueError):
            tag.insert_before(tag)

    def test_replace_with(self):
        soup = self.soup("<p>There's <b>no</b> business like <b>show</b> business</p>")
        no, show = soup.find_all("b")
        show.replace_with(no)
        assert soup.decode() == self.document_for(
            "<p>There's  business like <b>no</b> business</p>"
        )

        assert show.parent is None
        assert no.parent == soup.p
        assert no.next_element == "no"
        assert no.next_sibling == " business"

    def test_replace_with_errors(self):
        # Can't replace a tag that's not part of a tree.
        a_tag = Tag(name="a")
        with pytest.raises(ValueError):
            a_tag.replace_with("won't work")

        # Can't replace a tag with its parent.
        a_tag = self.soup("<a><b></b></a>").a
        with pytest.raises(ValueError):
            a_tag.b.replace_with(a_tag)

        # Or with a list that includes its parent.
        with pytest.raises(ValueError):
            a_tag.b.replace_with("string1", a_tag, "string2")

    def test_replace_with_multiple(self):
        data = "<a><b></b><c></c></a>"
        soup = self.soup(data)
        d_tag = soup.new_tag("d")
        d_tag.string = "Text In D Tag"
        e_tag = soup.new_tag("e")
        f_tag = soup.new_tag("f")
        a_string = "Random Text"
        soup.c.replace_with(d_tag, e_tag, a_string, f_tag)
        assert (
            soup.decode()
            == "<a><b></b><d>Text In D Tag</d><e></e>Random Text<f></f></a>"
        )
        assert soup.b.next_element == d_tag
        assert d_tag.string.next_element == e_tag
        assert e_tag.next_element.string == a_string
        assert e_tag.next_element.next_element == f_tag

    def test_replace_first_child(self):
        data = "<a><b></b><c></c></a>"
        soup = self.soup(data)
        soup.b.replace_with(soup.c)
        assert "<a><c></c></a>" == soup.decode()

    def test_replace_last_child(self):
        data = "<a><b></b><c></c></a>"
        soup = self.soup(data)
        soup.c.replace_with(soup.b)
        assert "<a><b></b></a>" == soup.decode()

    def test_nested_tag_replace_with(self):
        soup = self.soup(
            """<a>We<b>reserve<c>the</c><d>right</d></b></a><e>to<f>refuse</f><g>service</g></e>"""
        )

        # Replace the entire <b> tag and its contents ("reserve the
        # right") with the <f> tag ("refuse").
        remove_tag = soup.b
        move_tag = soup.f
        remove_tag.replace_with(move_tag)

        assert soup.decode() == self.document_for(
            "<a>We<f>refuse</f></a><e>to<g>service</g></e>"
        )

        # The <b> tag is now an orphan.
        assert remove_tag.parent is None
        assert remove_tag.find(string="right").next_element is None
        assert remove_tag.previous_element is None
        assert remove_tag.next_sibling is None
        assert remove_tag.previous_sibling is None

        # The <f> tag is now connected to the <a> tag.
        assert move_tag.parent == soup.a
        assert move_tag.previous_element == "We"
        assert move_tag.next_element.next_element == soup.e
        assert move_tag.next_sibling is None

        # The gap where the <f> tag used to be has been mended, and
        # the word "to" is now connected to the <g> tag.
        to_text = soup.find(string="to")
        g_tag = soup.g
        assert to_text.next_element == g_tag
        assert to_text.next_sibling == g_tag
        assert g_tag.previous_element == to_text
        assert g_tag.previous_sibling == to_text

    def test_unwrap(self):
        tree = self.soup("""
            <p>Unneeded <em>formatting</em> is unneeded</p>
            """)
        tree.em.unwrap()
        assert tree.em is None
        assert tree.p.text == "Unneeded formatting is unneeded"

    def test_wrap(self):
        soup = self.soup("I wish I was bold.")
        value = soup.string.wrap(soup.new_tag("b"))
        assert value.decode() == "<b>I wish I was bold.</b>"
        assert soup.decode() == self.document_for("<b>I wish I was bold.</b>")

    def test_wrap_extracts_tag_from_elsewhere(self):
        soup = self.soup("<b></b>I wish I was bold.")
        soup.b.next_sibling.wrap(soup.b)
        assert soup.decode() == self.document_for("<b>I wish I was bold.</b>")

    def test_wrap_puts_new_contents_at_the_end(self):
        soup = self.soup("<b>I like being bold.</b>I wish I was bold.")
        soup.b.next_sibling.wrap(soup.b)
        assert 2 == len(soup.b.contents)
        assert soup.decode() == self.document_for(
            "<b>I like being bold.I wish I was bold.</b>"
        )

    def test_extract(self):
        soup = self.soup(
            '<html><body>Some content. <div id="nav">Nav crap</div> More content.</body></html>'
        )

        assert len(soup.body.contents) == 3
        extracted = soup.find(id="nav").extract()

        assert soup.decode() == "<html><body>Some content.  More content.</body></html>"
        assert extracted.decode() == '<div id="nav">Nav crap</div>'

        # The extracted tag is now an orphan.
        assert len(soup.body.contents) == 2
        assert extracted.parent is None
        assert extracted.previous_element is None
        assert extracted.next_element.next_element is None

        # The gap where the extracted tag used to be has been mended.
        content_1 = soup.find(string="Some content. ")
        content_2 = soup.find(string=" More content.")
        assert content_1.next_element == content_2
        assert content_1.next_sibling == content_2
        assert content_2.previous_element == content_1
        assert content_2.previous_sibling == content_1

    def test_extract_distinguishes_between_identical_strings(self):
        soup = self.soup("<a>foo</a><b>bar</b>")
        foo_1 = soup.a.string
        foo_2 = soup.new_string("foo")
        bar_2 = soup.new_string("bar")
        soup.a.append(foo_2)
        soup.b.append(bar_2)

        # Now there are two identical strings in the <a> tag, and two
        # in the <b> tag. Let's remove the first "foo" and the second
        # "bar".
        foo_1.extract()
        bar_2.extract()
        assert foo_2 == soup.a.string
        assert bar_2 == soup.b.string

    def test_extract_multiples_of_same_tag(self):
        soup = self.soup("""
<html>
<head>
<script>foo</script>
</head>
<body>
 <script>bar</script>
 <a></a>
</body>
<script>baz</script>
</html>""")
        [soup.script.extract() for i in soup.find_all("script")]
        assert "<body>\n\n<a></a>\n</body>" == str(soup.body)

    def test_extract_works_when_element_is_surrounded_by_identical_strings(self):
        soup = self.soup("<html>\n" "<body>hi</body>\n" "</html>")
        soup.find("body").extract()
        assert None is soup.find("body")

    def test_clear(self):
        """Tag.clear()"""
        soup = self.soup("<p><a>String <em>Italicized</em></a> and another</p>")
        # clear using extract()
        a = soup.a
        soup.p.clear()
        assert len(soup.p.contents) == 0
        assert hasattr(a, "contents")

        # clear using decompose()
        em = a.em
        a.clear(decompose=True)
        assert 0 == len(em.contents)

    @pytest.mark.parametrize(
        "method_name,expected_result",
        [
            (
                "descendants",
                '<div><em>child1</em><p id="start"></p><p>child3</p></div>',
            ),
            (
                "next_siblings",
                '<div><em>child1</em><p id="start"><a>Second <em>child</em></a></p></div>',
            ),
            # Confused about why child3 is still here in this test? It's because removing the <p id="start"> tag from the tree removes all of its children from the tree as well. 'child'.next_element becomes None, because 'child' is no longer in the tree, and iteration stops there. Don't do this kind of thing, is what I'm saying.
            (
                "next_elements",
                '<div><em>child1</em><p id="start"></p><p>child3</p></div>',
            ),
            ("children", '<div><em>child1</em><p id="start"></p><p>child3</p></div>'),
            ("previous_elements", ""),
            (
                "previous_siblings",
                '<div><p id="start"><a>Second <em>child</em></a></p><p>child3</p></div>',
            ),
            ("parents", ""),
        ],
    )
    def test_extract_during_iteration(self, method_name, expected_result):
        # The iterators should be able to proceed even if the most
        # current yield got removed from the tree. This kind of code
        # is a bad idea, but we should be able to run it without an exception.
        soup = self.soup(
            "<div><em>child1</em><p id='start'><a>Second <em>child</em></a></p><p>child3</p></div>"
        )
        iterator = getattr(soup.p, method_name)
        for i in iterator:
            i.extract()
        assert expected_result == soup.decode()

    def test_decompose(self):
        # Test PageElement.decompose() and PageElement.decomposed
        soup = self.soup("<p><a>String <em>Italicized</em></a></p><p>Another para</p>")
        p1, p2 = soup.find_all("p")
        a = p1.a
        text = p1.em.string
        for i in [p1, p2, a, text]:
            assert False is i.decomposed

        # This sets p1 and everything beneath it to decomposed.
        p1.decompose()
        for i in [p1, a, text]:
            assert True is i.decomposed
        # p2 is unaffected.
        assert False is p2.decomposed

    def test_decompose_string(self):
        soup = self.soup("<div><p>String 1</p><p>String 2</p></p>")
        div = soup.div
        text = div.p.string
        assert False is text.decomposed
        text.decompose()
        assert True is text.decomposed
        assert "<div><p></p><p>String 2</p></div>" == div.decode()

    def test_string_set(self):
        """Tag.string = 'string'"""
        soup = self.soup("<a></a> <b><c></c></b>")
        soup.a.string = "foo"
        assert soup.a.contents == ["foo"]
        soup.b.string = "bar"
        assert soup.b.contents == ["bar"]

    def test_string_set_does_not_affect_original_string(self):
        soup = self.soup("<a><b>foo</b><c>bar</c>")
        soup.b.string = soup.c.string
        assert soup.a.encode() == b"<a><b>bar</b><c>bar</c></a>"

    def test_set_string_preserves_class_of_string(self):
        soup = self.soup("<a></a>")
        cdata = CData("foo")
        soup.a.string = cdata
        assert isinstance(soup.a.string, CData)


all_find_type_methods = [
    "find",
    "find_all",
    "find_parent",
    "find_parents",
    "find_next",
    "find_all_next",
    "find_previous",
    "find_all_previous",
    "find_next_sibling",
    "find_next_siblings",
    "find_previous_sibling",
    "find_previous_siblings",
]


class TestDeprecatedArguments(SoupTest):
    @pytest.mark.parametrize("method_name", all_find_type_methods)
    def test_find_type_method_string(self, method_name):
        soup = self.soup("<a>some</a><b>markup</b>")
        method = getattr(soup.b, method_name)
        with warnings.catch_warnings(record=True) as w:
            method(text="markup")
            [warning] = w
            assert warning.filename == __file__
            msg = str(warning.message)
            assert (
                msg
                == "The 'text' argument to find()-type methods is deprecated. Use 'string' instead."
            )


class TestWarnings(SoupTest):
    @pytest.mark.parametrize("method_name", all_find_type_methods)
    def test_suspicious_syntax_warning(self, method_name):
        soup = self.soup("<a>some</a><b>markup</b>")
        method = getattr(soup.b, method_name)
        with warnings.catch_warnings(record=True) as w:
            method(_class="u")
            [warning] = w
            assert warning.filename == __file__
            assert isinstance(warning.message, AttributeResemblesVariableWarning)
            msg = str(warning.message)
            assert (
                "'_class' is an unusual attribute name and is a common misspelling for 'class_'"
                in msg
            )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
    cast,
)

from playwright._impl._api_structures import FilePayload, FloatRect, Position
from playwright._impl._connection import ChannelOwner, from_nullable_channel
from playwright._impl._helper import (
    Error,
    KeyboardModifier,
    MouseButton,
    async_writefile,
    locals_to_params,
    make_dirs_for_file,
)
from playwright._impl._js_handle import (
    JSHandle,
    Serializable,
    parse_result,
    serialize_argument,
)
from playwright._impl._set_input_files_helpers import convert_input_files

if TYPE_CHECKING:  # pragma: no cover
    from playwright._impl._frame import Frame
    from playwright._impl._locator import Locator


class ElementHandle(JSHandle):
    def __init__(
        self, parent: ChannelOwner, type: str, guid: str, initializer: Dict
    ) -> None:
        super().__init__(parent, type, guid, initializer)
        self._frame = cast("Frame", parent)

    async def _createSelectorForTest(self, name: str) -> Optional[str]:
        return await self._channel.send(
            "createSelectorForTest", self._frame._timeout, dict(name=name)
        )

    def as_element(self) -> Optional["ElementHandle"]:
        return self

    async def owner_frame(self) -> Optional["Frame"]:
        return from_nullable_channel(await self._channel.send("ownerFrame", None))

    async def content_frame(self) -> Optional["Frame"]:
        return from_nullable_channel(await self._channel.send("contentFrame", None))

    async def get_attribute(self, name: str) -> Optional[str]:
        return await self._channel.send("getAttribute", None, dict(name=name))

    async def text_content(self) -> Optional[str]:
        return await self._channel.send("textContent", None)

    async def inner_text(self) -> str:
        return await self._channel.send("innerText", None)

    async def inner_html(self) -> str:
        return await self._channel.send("innerHTML", None)

    async def is_checked(self) -> bool:
        return await self._channel.send("isChecked", None)

    async def is_disabled(self) -> bool:
        return await self._channel.send("isDisabled", None)

    async def is_editable(self) -> bool:
        return await self._channel.send("isEditable", None)

    async def is_enabled(self) -> bool:
        return await self._channel.send("isEnabled", None)

    async def is_hidden(self) -> bool:
        return await self._channel.send("isHidden", None)

    async def is_visible(self) -> bool:
        return await self._channel.send("isVisible", None)

    async def dispatch_event(self, type: str, eventInit: Dict = None) -> None:
        await self._channel.send(
            "dispatchEvent",
            None,
            dict(type=type, eventInit=serialize_argument(eventInit)),
        )

    async def scroll_into_view_if_needed(self, timeout: float = None) -> None:
        await self._channel.send(
            "scrollIntoViewIfNeeded", self._frame._timeout, locals_to_params(locals())
        )

    async def hover(
        self,
        modifiers: Sequence[KeyboardModifier] = None,
        position: Position = None,
        timeout: float = None,
        noWaitAfter: bool = None,
        force: bool = None,
        trial: bool = None,
    ) -> None:
        await self._channel.send(
            "hover", self._frame._timeout, locals_to_params(locals())
        )

    async def click(
        self,
        modifiers: Sequence[KeyboardModifier] = None,
        position: Position = None,
        delay: float = None,
        button: MouseButton = None,
        clickCount: int = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
        trial: bool = None,
    ) -> None:
        await self._channel.send(
            "click", self._frame._timeout, locals_to_params(locals())
        )

    async def dblclick(
        self,
        modifiers: Sequence[KeyboardModifier] = None,
        position: Position = None,
        delay: float = None,
        button: MouseButton = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
        trial: bool = None,
    ) -> None:
        await self._channel.send(
            "dblclick", self._frame._timeout, locals_to_params(locals())
        )

    async def select_option(
        self,
        value: Union[str, Sequence[str]] = None,
        index: Union[int, Sequence[int]] = None,
        label: Union[str, Sequence[str]] = None,
        element: Union["ElementHandle", Sequence["ElementHandle"]] = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
    ) -> List[str]:
        params = locals_to_params(
            dict(
                timeout=timeout,
                force=force,
                **convert_select_option_values(value, index, label, element),
            )
        )
        return await self._channel.send("selectOption", self._frame._timeout, params)

    async def tap(
        self,
        modifiers: Sequence[KeyboardModifier] = None,
        position: Position = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
        trial: bool = None,
    ) -> None:
        await self._channel.send(
            "tap", self._frame._timeout, locals_to_params(locals())
        )

    async def fill(
        self,
        value: str,
        timeout: float = None,
        noWaitAfter: bool = None,
        force: bool = None,
    ) -> None:
        await self._channel.send(
            "fill", self._frame._timeout, locals_to_params(locals())
        )

    async def select_text(self, force: bool = None, timeout: float = None) -> None:
        await self._channel.send(
            "selectText", self._frame._timeout, locals_to_params(locals())
        )

    async def input_value(self, timeout: float = None) -> str:
        return await self._channel.send(
            "inputValue", self._frame._timeout, locals_to_params(locals())
        )

    async def set_input_files(
        self,
        files: Union[
            str, Path, FilePayload, Sequence[Union[str, Path]], Sequence[FilePayload]
        ],
        timeout: float = None,
        noWaitAfter: bool = None,
    ) -> None:
        frame = await self.owner_frame()
        if not frame:
            raise Error("Cannot set input files to detached element")
        converted = await convert_input_files(files, frame.page.context)
        await self._channel.send(
            "setInputFiles",
            self._frame._timeout,
            {
                "timeout": timeout,
                **converted,
            },
        )

    async def focus(self) -> None:
        await self._channel.send("focus", None)

    async def type(
        self,
        text: str,
        delay: float = None,
        timeout: float = None,
        noWaitAfter: bool = None,
    ) -> None:
        await self._channel.send(
            "type", self._frame._timeout, locals_to_params(locals())
        )

    async def press(
        self,
        key: str,
        delay: float = None,
        timeout: float = None,
        noWaitAfter: bool = None,
    ) -> None:
        await self._channel.send(
            "press", self._frame._timeout, locals_to_params(locals())
        )

    async def set_checked(
        self,
        checked: bool,
        position: Position = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
        trial: bool = None,
    ) -> None:
        if checked:
            await self.check(
                position=position,
                timeout=timeout,
                force=force,
                trial=trial,
            )
        else:
            await self.uncheck(
                position=position,
                timeout=timeout,
                force=force,
                trial=trial,
            )

    async def check(
        self,
        position: Position = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
        trial: bool = None,
    ) -> None:
        await self._channel.send(
            "check", self._frame._timeout, locals_to_params(locals())
        )

    async def uncheck(
        self,
        position: Position = None,
        timeout: float = None,
        force: bool = None,
        noWaitAfter: bool = None,
        trial: bool = None,
    ) -> None:
        await self._channel.send(
            "uncheck", self._frame._timeout, locals_to_params(locals())
        )

    async def bounding_box(self) -> Optional[FloatRect]:
        return await self._channel.send("boundingBox", None)

    async def screenshot(
        self,
        timeout: float = None,
        type: Literal["jpeg", "png"] = None,
        path: Union[str, Path] = None,
        quality: int = None,
        omitBackground: bool = None,
        animations: Literal["allow", "disabled"] = None,
        caret: Literal["hide", "initial"] = None,
        scale: Literal["css", "device"] = None,
        mask: Sequence["Locator"] = None,
        maskColor: str = None,
        style: str = None,
    ) -> bytes:
        params = locals_to_params(locals())
        if "path" in params:
            del params["path"]
        if "mask" in params:
            params["mask"] = list(
                map(
                    lambda locator: (
                        {
                            "frame": locator._frame._channel,
                            "selector": locator._selector,
                        }
                    ),
                    params["mask"],
                )
            )
        encoded_binary = await self._channel.send(
            "screenshot", self._frame._timeout, params
        )
        decoded_binary = base64.b64decode(encoded_binary)
        if path:
            make_dirs_for_file(path)
            await async_writefile(path, decoded_binary)
        return decoded_binary

    async def query_selector(self, selector: str) -> Optional["ElementHandle"]:
        return from_nullable_channel(
            await self._channel.send("querySelector", None, dict(selector=selector))
        )

    async def query_selector_all(self, selector: str) -> List["ElementHandle"]:
        return list(
            map(
                cast(Callable[[Any], Any], from_nullable_channel),
                await self._channel.send(
                    "querySelectorAll", None, dict(selector=selector)
                ),
            )
        )

    async def eval_on_selector(
        self,
        selector: str,
        expression: str,
        arg: Serializable = None,
    ) -> Any:
        return parse_result(
            await self._channel.send(
                "evalOnSelector",
                None,
                dict(
                    selector=selector,
                    expression=expression,
                    arg=serialize_argument(arg),
                ),
            )
        )

    async def eval_on_selector_all(
        self,
        selector: str,
        expression: str,
        arg: Serializable = None,
    ) -> Any:
        return parse_result(
            await self._channel.send(
                "evalOnSelectorAll",
                None,
                dict(
                    selector=selector,
                    expression=expression,
                    arg=serialize_argument(arg),
                ),
            )
        )

    async def wait_for_element_state(
        self,
        state: Literal[
            "disabled", "editable", "enabled", "hidden", "stable", "visible"
        ],
        timeout: float = None,
    ) -> None:
        await self._channel.send(
            "waitForElementState", self._frame._timeout, locals_to_params(locals())
        )

    async def wait_for_selector(
        self,
        selector: str,
        state: Literal["attached", "detached", "hidden", "visible"] = None,
        timeout: float = None,
        strict: bool = None,
    ) -> Optional["ElementHandle"]:
        return from_nullable_channel(
            await self._channel.send(
                "waitForSelector", self._frame._timeout, locals_to_params(locals())
            )
        )


def convert_select_option_values(
    value: Union[str, Sequence[str]] = None,
    index: Union[int, Sequence[int]] = None,
    label: Union[str, Sequence[str]] = None,
    element: Union["ElementHandle", Sequence["ElementHandle"]] = None,
) -> Any:
    if value is None and index is None and label is None and element is None:
        return {}

    options: Any = None
    elements: Any = None
    if value is not None:
        if isinstance(value, str):
            value = [value]
        options = (options or []) + list(map(lambda e: dict(valueOrLabel=e), value))
    if index is not None:
        if isinstance(index, int):
            index = [index]
        options = (options or []) + list(map(lambda e: dict(index=e), index))
    if label is not None:
        if isinstance(label, str):
            label = [label]
        options = (options or []) + list(map(lambda e: dict(label=e), label))
    if element:
        if isinstance(element, ElementHandle):
            element = [element]
        elements = list(map(lambda e: e._channel, element))

    return dict(options=options, elements=elements)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asyncio import AbstractEventLoop
from typing import Any, Optional

from playwright._impl._helper import Error
from playwright._impl._page import Page


class WebError:
    def __init__(
        self,
        loop: AbstractEventLoop,
        dispatcher_fiber: Any,
        page: Optional[Page],
        error: Error,
    ) -> None:
        self._loop = loop
        self._dispatcher_fiber = dispatcher_fiber
        self._page = page
        self._error = error

    @property
    def page(self) -> Optional[Page]:
        return self._page

    @property
    def error(self) -> Error:
        return self._error



#sofascore.py
from footystats.leagues import *
from footystats.webscrapping import *
from playwright.async_api import async_playwright
from playwright.async_api import Page
from footystats.webscrapping import LOCATIONS
import asyncio
#
from pathlib import Path
import json, pickle
import time
from datetime import datetime, timezone, timedelta

sofaroot:str="https://www.sofascore.com"

async def makeMatchRegister(x: League, path: str, current: bool) -> None:
    """
    Scrape match information for one or more leagues from Sofascore and
    save the results to disk as pickle files.

    Behavior:
      - Validates that `x` is an instance of the `Leagues` enum.
      - Ensures that the output directory `path` exists.
      - If `x == Leagues.ALL`, iterates through all available leagues 
        except the one with value 12.
      - For each target league:
          * Uses either `makeLeagueCurrentSeasonRegister` (if 
            `current=True`) or `makeLeaguePastSeasonRegister` (if 
            `current=False`) to gather match data.
          * Serializes the resulting dictionary into a `.pkl` file 
            named `<league>_match_register.pkl` inside the given path.

    Data organization in each pickle file:
      {
        "<league_name>": {
          "<season_name>": {
            "<round_name>": {
              ... match data from getMatchLinks() ...
            }
          }
        }
      }

    ⚠️ Notes:
      - Persistent browsing and random geolocation are applied inside
        the delegated league functions.
      - The `.pkl` files must be read with `pickle.load` when consumed.

    :param x: Target league enum. Use `Leagues.ALL` to scrape multiple 
              leagues in one run.
    :type x: League
    :param path: Filesystem path where results will be saved.
    :type path: str
    :param current: If True, scrape the current season. If False, scrape 
                    past seasons.
    :type current: bool

    :raises TypeError: If `x` is not an instance of `Leagues`.
    :raises FileNotFoundError: If the given `path` does not exist.

    :return: None. Results are saved to pickle files.
    :rtype: None
    """
    if not isinstance(x,Leagues):
        raise TypeError("Expected instance Leagues(Enum)")
    if not Path(path).exists():
        raise FileNotFoundError(f"Path does not exist: {path}")
    if x==Leagues.ALL:
        for league in list(filter(lambda x : x.value != 12, Leagues)):
            if current:
                data = await makeLeagueCurrentSeasonRegister(league)
            else:
                data = await makeLeaguePastSeasonRegister(league)
            leagueName:str=league.name.lower()
            filename = leagueName+'_match_register.pkl'
            p=Path(path).joinpath(path,filename)
            with open(p,"wb") as f:
                pickle.dump(data,f)
            f.close()
    else:
        if current:
            data = await makeLeagueCurrentSeasonRegister(league)
        else:
            data = await makeLeaguePastSeasonRegister(league)
        leagueName:str=x.name.lower()
        filename = leagueName+'_match_register.pkl'
        p=Path(path).joinpath(path,filename)
        print(p)
        with open(p,"wb") as f:
            pickle.dump(data,f)
        f.close()
        
async def makeLeaguePastSeasonRegister(x:League)->dict:
    """
    Scrape match information for past seasons of a given league from 
    its Sofascore page.

    The function:
      - Simulates browsing from a randomly chosen world city 
        (via spoofed geolocation and locale).
      - Opens the Sofascore league page.
      - Iterates through a set of past seasons (currently the 2 most 
        recent ones, excluding the current).
      - For each season:
          * Detects whether a "rounds" dropdown exists.
          * If rounds exist, iterates through all available rounds.
          * Filters out non-standard competitions 
            (qualification, relegation, playoff).
          * Collects match URLs for valid rounds using `getMatchLinks`.

    Data is structured hierarchically in a nested dictionary, grouped by:
      - league name
      - season
      - round

    ⚠️ Notes:
      - Only rounds explicitly containing the word "round" are included.
      - Seasons are currently restricted to indices 1 and 2 
        (`range(1, 3)`), meaning the function processes the two most 
        recent past seasons. Adjust the loop to cover more if needed.
      - Uses a persistent browser profile (`./my_profile`) so cookies 
        and settings are retained across runs.
      - If dropdown elements cannot be found or interacted with, 
        the function skips them and continues gracefully.

    :param x: Enum value representing the league whose matches 
              should be scraped.
    :type x: League

    :raises RuntimeError: if critical dropdowns (season or round) 
                          cannot be located or interacted with.

    :return: Nested dictionary of scraped match URLs and metadata.
             Example:
             {
               "bundesliga": {
                 "2022/2023": {
                   "Round 1": {
                     ... match data from getMatchLinks() ...
                   },
                   "Round 2": { ... },
                   ...
                 },
                 "2021/2022": { ... }
               }
             }
    :rtype: dict
    """
    #
    cLeague     = leagues[x.value]
    # filename    = x.name.lower()+"_sources.json"
    # savepath    = Path(path).joinpath(filename)
    city = random.choice(list(LOCATIONS.keys()))  # Randomly pick a city each run
    location = LOCATIONS[city]
    print(f"[+] Simulating location: {city}")
    #
    async with async_playwright() as p:
        browser = await p.chromium.launch_persistent_context(
        user_data_dir="./my_profile",
        headless=False,
        args=["--start-maximized",
        # f"--disable-extensions-except={path_to_extension}",
        # f"--load-extension={path_to_extension}"
        ],
        viewport=None,  # Forces full screen
        locale=location["locale"],  # default
        geolocation={"latitude": location["latitude"], "longitude": location["longitude"]},
        permissions=["geolocation"],
        
        )
        page = await browser.new_page()
        # await rotate_locations(browser, page, delay=90)

        # ➤ Stealth: remove webdriver flag
        await page.add_init_script("Object.defineProperty(navigator, 'webdriver', {get: () => undefined})")
        catalog:dict={}
        # ==================================================================
        # (0) LOAD MAIN PAGE
        # ==================================================================
        await block_ads(page)
        await page.goto(cLeague.sofa_url)
        # await page.mouse.wheel(random.uniform(100, 550), random.uniform(900, 1500))
        #
        # ==================================================================
        # SELECT SEASON
        # ==================================================================
        # 1. Locate the seasons dropdown container and open its combobox
        season_dropdown = page.locator('div.Dropdown.kdhXwd')
        season_combobox = season_dropdown.locator('button[role="combobox"]')
        await season_combobox.click()
        # 2. Get all season options inside that container
        season_options = season_dropdown.locator('ul[role="listbox"] li')
        season_count = await season_options.count()
        # ==================================================================
        # FOR EACH SEASON...
        # ==================================================================
        for i in range(1,6):
            # await asyncio.sleep(100)
            season_text = await season_options.nth(i).inner_text()
            if i==1:
                catalog[cLeague.name.lower()]={}
            catalog[cLeague.name.lower()][season_text]={}
            print(f"Clicking season {i+1}: {season_text}")
            await season_options.nth(i).click()
            await page.wait_for_load_state("networkidle")
            await asyncio.sleep(1.0)
            await page.evaluate("window.scrollTo(0, 0)")
            # ===============================================================
            # DETECTION OF ROUNDS
            # ===============================================================
            round_dropdown = page.locator('div[data-panelid="round"] div.Dropdown.gSFIyj')
            if await round_dropdown.count() > 0:
                # ===========================================================
                # ROUNDS FOUND, LETS GO...
                # ===========================================================
                await round_dropdown.wait_for(state="attached")
                # Open the round combobox
                round_combobox = round_dropdown.locator('button[role="combobox"]').first
                await round_combobox.wait_for(state="attached")
                await round_combobox.click()
                await page.wait_for_load_state("networkidle")
                # Get round options
                round_options = round_dropdown.locator('ul[role="listbox"] li')
                await round_options.first.wait_for(state='attached')
                round_count = await round_options.count()
                for j in range(round_count):
                    round_combobox = round_dropdown.locator('button[role="combobox"]').first
                    await round_combobox.wait_for(state="attached")
                    round_text = await round_options.nth(j).inner_text()
                    print(f"   Clicking round {j+1}: {round_text}")
                    
                    await round_options.nth(j).click()
                    # await page.wait_for_load_state("networkidle")
                    print("      ✅ Round clicked")
                    await round_combobox.click()
                    # ======================================================
                    # GRAB MATCHES URLS
                    # ======================================================
                    isRound = round_text.lower().find('round')!=-1
                    isQualification = round_text.lower().find('qualification')!=-1
                    isRelegation = round_text.lower().find('relegation')!=-1
                    isPlayoff = round_text.lower().find('playoff')!=-1
                    if isRound and not isQualification and not isRelegation and not isPlayoff:
                        data:dict = await getMatchLinks(page)
                        catalog[cLeague.name.lower()][season_text][round_text]=data
                    
            # ==============================================================
            # NO ROUNDS BUTTON THEN SKIP
            # ==============================================================
            else:
                print(f"   NO ROUNDS: SKIP SEASON {season_text}")
            # ==============================================================
            # SEASON COMPLETE
            # ==============================================================
            if i < season_count - 1:
                season_dropdown = page.locator('div.Dropdown.kdhXwd')
                season_combobox = season_dropdown.locator('button[role="combobox"]')
                await season_combobox.wait_for(state="visible")
                await season_combobox.click()
            await asyncio.sleep(1.0)
        await browser.close()
        return dict(catalog)

async def makeLeagueCurrentSeasonRegister(x:League)->dict:
    """
    Scrape match information for the current season of a given league 
    from its Sofascore page.

    The function:
      - Simulates browsing from a randomly chosen world city 
        (using spoofed geolocation and locale).
      - Navigates to the league's Sofascore page.
      - Iterates through available seasons (currently limited to the most 
        recent one).
      - For each season:
          * Detects whether a "rounds" dropdown exists.
          * If rounds exist, iterates through all available rounds.
          * Filters out non-standard competitions 
            (e.g. qualification, relegation, playoff).
          * Collects match URLs for valid rounds via `getMatchLinks`.

    Data is structured hierarchically in a nested dictionary, grouped by:
      - league name
      - season
      - round

    ⚠️ Notes:
      - Only rounds explicitly containing the word "round" are included.
      - The browser is launched in persistent mode with a dedicated 
        profile folder, so cookies and browsing data persist across runs.
      - If dropdown elements cannot be found or interacted with, the 
        function skips them and continues.

    :param x: Enum value representing the league whose matches 
              should be scraped.
    :type x: League

    :raises RuntimeError: if critical dropdowns (season or round) 
                          cannot be located or interacted with.

    :return: Nested dictionary of scraped match URLs and metadata.
             Example:
             {
               "premier_league": {
                 "2023/2024": {
                   "Round 1": {
                     ... match data from getMatchLinks() ...
                   },
                   "Round 2": { ... },
                   ...
                 }
               }
             }
    :rtype: dict
    """
    #
    cLeague     = leagues[x.value]
    # filename    = x.name.lower()+"_sources.json"
    # savepath    = Path(path).joinpath(filename)
    city = random.choice(list(LOCATIONS.keys()))  # Randomly pick a city each run
    location = LOCATIONS[city]
    print(f"[+] Simulating location: {city}")
    #
    async with async_playwright() as p:
        browser = await p.chromium.launch_persistent_context(
        user_data_dir="./my_profile",
        headless=False,
        args=["--start-maximized",
        # f"--disable-extensions-except={path_to_extension}",
        # f"--load-extension={path_to_extension}"
        ],
        viewport=None,  # Forces full screen
        locale=location["locale"],  # default
        geolocation={"latitude": location["latitude"], "longitude": location["longitude"]},
        permissions=["geolocation"],
        
        )
        page = await browser.new_page()
        # await rotate_locations(browser, page, delay=90)

        # ➤ Stealth: remove webdriver flag
        await page.add_init_script("Object.defineProperty(navigator, 'webdriver', {get: () => undefined})")
        catalog:dict={}
        # ==================================================================
        # (0) LOAD MAIN PAGE
        # ==================================================================
        await block_ads(page)
        await page.goto(cLeague.sofa_url)
        # await page.mouse.wheel(random.uniform(100, 550), random.uniform(900, 1500))
        #
        # ==================================================================
        # SELECT SEASON
        # ==================================================================
        # 1. Locate the seasons dropdown container and open its combobox
        season_dropdown = page.locator('div.Dropdown.kdhXwd')
        season_combobox = season_dropdown.locator('button[role="combobox"]')
        await season_combobox.click()
        # 2. Get all season options inside that container
        season_options = season_dropdown.locator('ul[role="listbox"] li')
        season_count = await season_options.count()
        # ==================================================================
        # FOR EACH SEASON...
        # ==================================================================
        for i in range(0,1):
            # await asyncio.sleep(100)
            season_text = await season_options.nth(i).inner_text()
            catalog[cLeague.name.lower()]={}
            catalog[cLeague.name.lower()][season_text]={}
            print(f"Clicking season {i+1}: {season_text}")
            await season_options.nth(i).click()
            await page.wait_for_load_state("networkidle")
            await asyncio.sleep(1.0)
            await page.evaluate("window.scrollTo(0, 0)")
            # ===============================================================
            # DETECTION OF ROUNDS
            # ===============================================================
            round_dropdown = page.locator('div[data-panelid="round"] div.Dropdown.gSFIyj')
            if await round_dropdown.count() > 0:
                # ===========================================================
                # ROUNDS FOUND, LETS GO...
                # ===========================================================
                await round_dropdown.wait_for(state="attached")
                # Open the round combobox
                round_combobox = round_dropdown.locator('button[role="combobox"]').first
                await round_combobox.wait_for(state="attached")
                await round_combobox.click()
                await page.wait_for_load_state("networkidle")
                # Get round options
                round_options = round_dropdown.locator('ul[role="listbox"] li')
                await round_options.first.wait_for(state='attached')
                round_count = await round_options.count()
                for j in range(round_count):
                    round_combobox = round_dropdown.locator('button[role="combobox"]').first
                    await round_combobox.wait_for(state="attached")
                    round_text = await round_options.nth(j).inner_text()
                    print(f"   Clicking round {j+1}: {round_text}")
                    
                    await round_options.nth(j).click()
                    # await page.wait_for_load_state("networkidle")
                    print("      ✅ Round clicked")
                    await round_combobox.click()
                    # ======================================================
                    # GRAB MATCHES URLS
                    # ======================================================
                    isRound = round_text.lower().find('round')!=-1
                    isQualification = round_text.lower().find('qualification')!=-1
                    isRelegation = round_text.lower().find('relegation')!=-1
                    isPlayoff = round_text.lower().find('playoff')!=-1
                    if isRound and not isQualification and not isRelegation and not isPlayoff:
                        data:dict = await getMatchLinks(page)
                        catalog[cLeague.name.lower()][season_text][round_text]=data
                    
            # ==============================================================
            # NO ROUNDS BUTTON THEN SKIP
            # ==============================================================
            else:
                print(f"   NO ROUNDS: SKIP SEASON {season_text}")
            # ==============================================================
            # SEASON COMPLETE
            # ==============================================================
            if i < season_count - 1:
                season_dropdown = page.locator('div.Dropdown.kdhXwd')
                season_combobox = season_dropdown.locator('button[role="combobox"]')
                await season_combobox.wait_for(state="visible")
                await season_combobox.click()
            await asyncio.sleep(1.0)

        await browser.close()
        return dict(catalog)

async def getMatchLinks(page:Page)->dict:
    """
    scrape match information from a Sofascore page. the function 
    extracts home and away team names, final scores, match URLs, 
    and results (win/loss/draw), and returns them as a dictionary.

    if a team name is not found, the function raises an error. 
    only matches with available scores and status information 
    are included in the output.

    :param page: playwright Page object of the Sofascore webpage 
                 being scraped
    :type page: Page

    :raises EmptyTextError: if the home or away team label cannot 
                            be found on the page

    :return: dictionary containing lists of scraped match data with:
             - 'hometeam': list of home team names
             - 'awayteam': list of away team names
             - 'url': list of Sofascore match URLs
             - 'result': list of results ('W', 'L', or 'D')
    :rtype: dict
    """
    data={}
    data['hometeam']=[]
    data['awayteam']=[]
    data['url']=[]
    data['result']=[]
    #
    matches_box = page.locator("div.Box.kiSsvW")
    await matches_box.wait_for(state="attached")
    links = matches_box.locator("a")
    link_count = await links.count()
    for i in range(link_count):
        href = await links.nth(i).get_attribute("href")
        print(sofaroot+href)
        #
        trueMatch = href.find('tournament')==-1
        if trueMatch:
            hometeam=""
            awayteam=""
            hometeamLabelLocator = links.nth(i).locator("div.Box.Flex.khxXvq.jLRkRA")
            awayteamLabelLocator = links.nth(i).locator("div.Box.Flex.ggRYVx.jLRkRA")
            # ==============================================================
            # get HOMETEAM LABEL
            # ==============================================================
            bdi_home = hometeamLabelLocator.nth(0).locator("bdi.Text.ezSveL")
            bdi_home_count = await bdi_home.count()
            if bdi_home_count:
                hometeam = await bdi_home.nth(0).inner_text()
            else:
                bdi_home = hometeamLabelLocator.nth(0).locator("bdi.Text.kwIkWN")
                hometeam = await bdi_home.nth(0).inner_text()
            # ==============================================================
            # get AWAYTEAM LABEL
            # ==============================================================                
            bdi_away = awayteamLabelLocator.nth(1).locator("bdi.Text.ezSveL")
            bdi_away_count = await bdi_away.count()
            if bdi_away_count:
                awayteam = await bdi_away.nth(0).inner_text()
            else:
                bdi_away = awayteamLabelLocator.nth(1).locator("bdi.Text.kwIkWN")
                awayteam = await bdi_away.nth(0).inner_text()
            # ===================================================
            if hometeam =="": 
                raise EmptyTextError(f"ERROR SCRAPPING MATCH:\n{href}\n HOME team is not foun")
            if awayteam =="": 
                raise EmptyTextError(f"ERROR SCRAPPING MATCH:\n{href}\n AWAY team is not found")
            # ==============================================================
            # get SCORES
            # ==============================================================                        
            score_box_locator = links.nth(i).locator("div.Box.Flex.btcITE.yaNbA")
            inner_score_box = score_box_locator.locator("div.Box.Flex.jTiCHC.MkeW.sc-37b4466f-2.fUqQjG.score-box")
            
            homescore=""
            awayscore=""
            
            forbiddens = ['\n0','\n1','\n2','\n3','\n4','\n5','\n6','\n7','\n8']
            replacement = lambda text: (lambda t: [t := t.replace(f, '') for f in forbiddens] and t)(text)

            if await inner_score_box.count():
                nb_score_box=await inner_score_box.count()
                homescore = await inner_score_box.nth(0).inner_text()
                awayscore = await inner_score_box.nth(1).inner_text()
                homescore=replacement(homescore)
                awayscore=replacement(awayscore)
            # ==============================================================
            # get MATCH STATUS
            # ============================================================== 
            ft_locator = links.nth(i).locator("span.Text.fjeMtb.currentScore bdi.Text.kkVniA")
            ft_text="None"
            if await ft_locator.count():
                ft_text = await ft_locator.inner_text()
            print(f"Match status: {ft_text}")
            if homescore!="" and awayscore!="" and ft_text!="None":
                data['hometeam'].append(hometeam)
                data['awayteam'].append(awayteam)
                data['url'].append(sofaroot+href)
                if int(homescore)>int(awayscore):
                    data['result'].append('W')
                if int(awayscore)>int(homescore):
                    data['result'].append('L')
                if int(awayscore)==int(homescore):
                    data['result'].append('D')
                # dictToUpdate[
            print(f"HOMETEAM: {hometeam} {homescore}")
            print(f"AWAYTEAM: {awayteam} {awayscore}")
    return dict(data)

async def parseMatch(matchurl:str):
    """
    """
    pass
    

class EmptyTextError(Exception):
    """Raised when a locator returns empty text unexpectedly."""
    pass


#whoscored.py
import asyncio
from pathlib import Path
import random
from playwright.async_api import async_playwright
import json
import time
from datetime import datetime, timezone, timedelta
from footystats.leagues import *
from footystats.webscrapping import block_ads

async def goToLeague(x:League, path:str):
    """
    """
    #
    cLeague     = leagues[x.value]
    filename    = x.name.lower()+"_sources.json"
    savepath    = Path(path).joinpath(filename)
    #
    async with async_playwright() as p:
        browser = await p.chromium.launch_persistent_context(
        user_data_dir="./my_profile",
        headless=False,
        args=["--start-maximized",
        # f"--disable-extensions-except={path_to_extension}",
        # f"--load-extension={path_to_extension}"
        ],
        viewport=None,  # Forces full screen
        )
        page = await browser.new_page()

        # ➤ Stealth: remove webdriver flag
        await page.add_init_script("Object.defineProperty(navigator, 'webdriver', {get: () => undefined})")
        
        # ==================================================================
        # (0) LOAD MAIN PAGE
        # ==================================================================
        # await block_ads(page)
        await page.goto(cLeague.ws_url)
        # await page.mouse.wheel(random.uniform(100, 550), random.uniform(900, 1500))
        
        html_content = await page.content()
        listbox = soup.find("ul", {"role": "listbox", "class" : "Box klGMtt"})
        
        for ul in soup.find_all('ul' , {"role": "listbox"}):
            print('toto')

        # if listbox:
            # print("Found listbox:")
            # print(listbox.prettify())
        # else:
            # print("No listbox found.")
            
        # if listbox:
        # # Find all <li> inside with role="option"
            # options = listbox.find_all("li", {"role": "option"})
            
            # for opt in options:
                print("Value:", opt.get("value"), "| Text:", opt.get_text(strip=True))
        
        await asyncio.sleep(20)


# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import io
import json
import os
import subprocess
import sys
from abc import ABC, abstractmethod
from typing import Callable, Dict, Optional, Union

from playwright._impl._driver import compute_driver_executable, get_driver_env
from playwright._impl._helper import ParsedMessagePayload


# Sourced from: https://github.com/pytest-dev/pytest/blob/da01ee0a4bb0af780167ecd228ab3ad249511302/src/_pytest/faulthandler.py#L69-L77
def _get_stderr_fileno() -> Optional[int]:
    try:
        # when using pythonw, sys.stderr is None.
        # when Pyinstaller is used, there is no closed attribute because Pyinstaller monkey-patches it with a NullWriter class
        if sys.stderr is None or not hasattr(sys.stderr, "closed"):
            return None
        if sys.stderr.closed:
            return None

        return sys.stderr.fileno()
    except (NotImplementedError, AttributeError, io.UnsupportedOperation):
        # pytest-xdist monkeypatches sys.stderr with an object that is not an actual file.
        # https://docs.python.org/3/library/faulthandler.html#issue-with-file-descriptors
        # This is potentially dangerous, but the best we can do.
        if not hasattr(sys, "__stderr__") or not sys.__stderr__:
            return None
        return sys.__stderr__.fileno()


class Transport(ABC):
    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self._loop = loop
        self.on_message: Callable[[ParsedMessagePayload], None] = lambda _: None
        self.on_error_future: asyncio.Future = loop.create_future()

    @abstractmethod
    def request_stop(self) -> None:
        pass

    def dispose(self) -> None:
        pass

    @abstractmethod
    async def wait_until_stopped(self) -> None:
        pass

    @abstractmethod
    async def connect(self) -> None:
        pass

    @abstractmethod
    async def run(self) -> None:
        pass

    @abstractmethod
    def send(self, message: Dict) -> None:
        pass

    def serialize_message(self, message: Dict) -> bytes:
        msg = json.dumps(message)
        if "DEBUGP" in os.environ:  # pragma: no cover
            print("\x1b[32mSEND>\x1b[0m", json.dumps(message, indent=2))
        return msg.encode()

    def deserialize_message(self, data: Union[str, bytes]) -> ParsedMessagePayload:
        obj = json.loads(data)

        if "DEBUGP" in os.environ:  # pragma: no cover
            print("\x1b[33mRECV>\x1b[0m", json.dumps(obj, indent=2))
        return obj


class PipeTransport(Transport):
    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        super().__init__(loop)
        self._stopped = False

    def request_stop(self) -> None:
        assert self._output
        self._stopped = True
        self._output.close()

    async def wait_until_stopped(self) -> None:
        await self._stopped_future

    async def connect(self) -> None:
        self._stopped_future: asyncio.Future = asyncio.Future()

        try:
            # For pyinstaller and Nuitka
            env = get_driver_env()
            if getattr(sys, "frozen", False) or globals().get("__compiled__"):
                env.setdefault("PLAYWRIGHT_BROWSERS_PATH", "0")

            startupinfo = None
            if sys.platform == "win32":
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                startupinfo.wShowWindow = subprocess.SW_HIDE

            executable_path, entrypoint_path = compute_driver_executable()
            self._proc = await asyncio.create_subprocess_exec(
                executable_path,
                entrypoint_path,
                "run-driver",
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=_get_stderr_fileno(),
                limit=32768,
                env=env,
                startupinfo=startupinfo,
            )
        except Exception as exc:
            self.on_error_future.set_exception(exc)
            raise exc

        self._output = self._proc.stdin

    async def run(self) -> None:
        assert self._proc.stdout
        assert self._proc.stdin
        while not self._stopped:
            try:
                buffer = await self._proc.stdout.readexactly(4)
                if self._stopped:
                    break
                length = int.from_bytes(buffer, byteorder="little", signed=False)
                buffer = bytes(0)
                while length:
                    to_read = min(length, 32768)
                    data = await self._proc.stdout.readexactly(to_read)
                    if self._stopped:
                        break
                    length -= to_read
                    if len(buffer):
                        buffer = buffer + data
                    else:
                        buffer = data
                if self._stopped:
                    break

                obj = self.deserialize_message(buffer)
                self.on_message(obj)
            except asyncio.IncompleteReadError:
                if not self._stopped:
                    self.on_error_future.set_exception(
                        Exception("Connection closed while reading from the driver")
                    )
                break
            await asyncio.sleep(0)

        await self._proc.communicate()
        self._stopped_future.set_result(None)

    def send(self, message: Dict) -> None:
        assert self._output
        data = self.serialize_message(message)
        self._output.write(
            len(data).to_bytes(4, byteorder="little", signed=False) + data
        )
